// generate LUControl module
// by Wei Zhang
//

#include <stdio.h>
#include <math.h>
#define intlog2(x) (int)ceil(log(x)/log(2))

void genBlockControl(int numPE, int precision, int mwidth, int mdivk, int intdivlat, int multlat, int addlat,
    int curlat, int leftlat, int toplat, int divlat, int ramwidth, int ramsizewidth, int topsizewidth)
{
    FILE *fp;
    int PEwidth, mdivkwidth;
    int maxcurleftlat, maxtopleftlat, maxcurtoplat, maxmemlat;
    int divlatwidth;
    int madelay, mcdelay, mdelay, ltdelay;
    int i;

    maxtopleftlat = (toplat > leftlat) ? toplat : leftlat;
    maxcurleftlat = (curlat > leftlat) ? curlat : leftlat;
    maxcurtoplat = (curlat > toplat) ? curlat : toplat;
    maxmemlat = (maxcurtoplat > leftlat) ? maxcurtoplat : leftlat;
    PEwidth = intlog2(numPE);
    divlat += toplat;
    divlatwidth = intlog2(divlat+1);
    mdivkwidth = intlog2(mdivk+1);
    madelay = multlat+addlat+maxtopleftlat+1;
    mdelay = multlat+maxtopleftlat-curlat;
    mcdelay = multlat+leftlat;
    ltdelay = abs(toplat-leftlat);
    //cdelay = leftlat-toplat;

    fp = fopen("LUControl.v", "w");

    fprintf(fp, "//auto-generated LUControl.v\n");
    fprintf(fp, "//control block that creates all the control signals\n");
    fprintf(fp, "//by Wei Zhang\n\n");
    fprintf(fp, "module LUControl (clk, start_in, m_in, n_in, loop_in, mode_in, done,\n");
    fprintf(fp, "					curReadAddr, curWriteAddr, curWriteByteEn, curWriteEn, curWriteSel, \n");
    fprintf(fp, "					leftReadAddr, leftWriteAddr, leftWriteByteEn, leftWriteEn,  leftWriteSel,\n");
    fprintf(fp, "					topReadAddr, topWriteAddr, topWriteEn, topWriteSel, topSourceSel, diagEn, MOSel, MOEn);\n");
    fprintf(fp, "\n");
    fprintf(fp, "parameter NUMPE = %i, PEWIDTH = %i;\n", numPE, PEwidth);
    fprintf(fp, "parameter BLOCKWIDTH = %i, BLOCKDIVK = %i, BLOCKDIVKWIDTH = %i;\n", mwidth, mdivk, mdivkwidth);
    fprintf(fp, "parameter INTDIVLAT = %i;\n", intdivlat);
    fprintf(fp, "parameter CURLAT = %i, LEFTLAT = %i, TOPLAT = %i, DIVLAT = %i;\n", curlat, leftlat, toplat, divlat);
    fprintf(fp, "parameter DIVLATWIDTH = %i, MADELAY = %i, MCDELAY = %i, MDELAY = %i;\n", divlatwidth, madelay, mcdelay, mdelay);
    fprintf(fp, "parameter LTDELAY = %i;\n", ltdelay);
    fprintf(fp, "parameter MAXCURLEFTLAT = %i, MAXTOPLEFTLAT = %i, MAXCURTOPLAT = %i;\n", maxcurleftlat, maxtopleftlat, maxcurtoplat);
    fprintf(fp, "parameter RAMNUMBYTES = %i, RAMSIZEWIDTH = %i, TOPSIZEWIDTH = %i;\n", ramwidth/8, ramsizewidth, topsizewidth);
    fprintf(fp, "\n");
    fprintf(fp, "input clk, start_in;\n");
    fprintf(fp, "input[BLOCKWIDTH-1:0] m_in, n_in, loop_in;\n");
    fprintf(fp, "input[1:0] mode_in;\n");
    fprintf(fp, "output done;\n");
    fprintf(fp, "\n");
    fprintf(fp, "output[RAMNUMBYTES-1:0] curWriteByteEn;\n");
    fprintf(fp, "output[RAMSIZEWIDTH-1:0] curWriteAddr, curReadAddr;\n");
    fprintf(fp, "output curWriteEn;\n");
    fprintf(fp, "\n");
    fprintf(fp, "output[RAMNUMBYTES-1:0] leftWriteByteEn;\n");
    fprintf(fp, "output[RAMSIZEWIDTH-1:0] leftWriteAddr, leftReadAddr;\n");
    fprintf(fp, "output leftWriteEn;\n");
    fprintf(fp, "\n");
    fprintf(fp, "output[TOPSIZEWIDTH-1:0] topWriteAddr, topReadAddr;\n");
    fprintf(fp, "output topWriteEn;\n");
    fprintf(fp, "\n");
    fprintf(fp, "output leftWriteSel, curWriteSel, topSourceSel, diagEn;\n");
    fprintf(fp, "output[PEWIDTH-1:0] topWriteSel;\n");
    fprintf(fp, "\n");
    fprintf(fp, "output MOSel;\n");
    fprintf(fp, "output MOEn;\n");
    fprintf(fp, "\n");
    fprintf(fp, "parameter SETUP = 0, START = 1, FETCH_COL = 2, WAIT_COL = 3, FIND_REC = 4, MULT_COL = 5, UPDATE_J = 6,\n");
    fprintf(fp, "		  STORE_MO = 7, MULT_SUB = 8, INCRE_I = 9, WAIT = 10, DONE = 11,  STORE_DIAG = 12, STORE_DIAG2 = 13, START_FETCH_ROW = 14;\n");
    fprintf(fp, "parameter ROW_WAIT = 0, FETCH_ROW = 1, DONE_FETCH_ROW = 2, LOAD_ROW_INC_J = 3;\n");
    fprintf(fp, "\n");
    fprintf(fp, "reg start, startDelay[INTDIVLAT:0];\n");
    fprintf(fp, "reg[BLOCKWIDTH-1:0] m, n, stop, stop2, loop;\n");
    fprintf(fp, "reg[1:0] mode;\n");
    fprintf(fp, "reg[3:0] nextState, currentState;\n");
    fprintf(fp, "reg[1:0] nextRowState, currentRowState;\n");
    fprintf(fp, "reg startFetchRow, doneFetchRow, loadRow, writeRow;\n");
    fprintf(fp, "reg updateCounter;\n");
    fprintf(fp, "\n");
    fprintf(fp, "reg[BLOCKWIDTH-1:0] i1, j;\n");
    fprintf(fp, "reg[TOPSIZEWIDTH-1:0] nextTopIdx, nextTopIdx2, curTopIdx, nextTopIdxCounter;\n");
    fprintf(fp, "reg[BLOCKDIVKWIDTH-1:0] topIdx, topIdxCounter, mdivk;\n");
    fprintf(fp, "reg[RAMSIZEWIDTH-1:0] diagIdx, leftIdx, msIdx;\n");
    fprintf(fp, "reg[PEWIDTH-1:0] imodk, i1modk;\n");
    fprintf(fp, "reg[RAMSIZEWIDTH-1:0] diagIdxCounter, leftIdxCounter, msIdxCounter, readRowCounter, topWriteCounter;\n");
    fprintf(fp, "reg[RAMNUMBYTES-1:0] byteEn, i1modkByteEn;\n");
    fprintf(fp, "\n");
    fprintf(fp, "reg done;\n");
    fprintf(fp, "\n");
    fprintf(fp, "reg[RAMNUMBYTES-1:0] curWriteByteEn;\n");
    fprintf(fp, "reg[RAMSIZEWIDTH-1:0] curWriteAddr, curReadAddr;\n");
    fprintf(fp, "reg curWriteEn;\n");
    fprintf(fp, "\n");
    fprintf(fp, "reg[RAMNUMBYTES-1:0] leftWriteByteEn;\n");
    fprintf(fp, "reg[RAMSIZEWIDTH-1:0] leftWriteAddr, leftReadAddr;\n");
    fprintf(fp, "reg leftWriteEn;\n");
    fprintf(fp, "\n");
    fprintf(fp, "reg[TOPSIZEWIDTH-1:0] topWriteAddr, topReadAddr;\n");
    fprintf(fp, "reg topWriteEn;\n");
    fprintf(fp, "\n");
    fprintf(fp, "reg leftWriteSel, curWriteSel, topSourceSel, diagEn;\n");
    fprintf(fp, "reg[PEWIDTH-1:0] topWriteSel;\n");
    fprintf(fp, "\n");
    fprintf(fp, "reg MOSel;\n");
    fprintf(fp, "reg MOEn;\n");
    fprintf(fp, "\n");
    fprintf(fp, "reg[RAMSIZEWIDTH-1:0] counter;\n");
    fprintf(fp, "reg[DIVLATWIDTH-1:0] divCounter;\n");
    fprintf(fp, "\n");
    fprintf(fp, "reg[RAMNUMBYTES-1:0] writeByteEnDelay[MADELAY-1:0];\n");
    fprintf(fp, "reg[RAMSIZEWIDTH-1:0] curWriteAddrDelay[MADELAY-1:0], curReadAddrDelay[MDELAY-1:0];\n");
    fprintf(fp, "reg leftWriteEnDelay[MADELAY-1:0], curWriteEnDelay[MADELAY-1:0];\n");
    fprintf(fp, "reg leftWriteSelDelay[CURLAT-1:0], curWriteSelDelay[MCDELAY-1:0];\n");
    if (toplat > leftlat) {
        fprintf(fp, "reg[RAMSIZEWIDTH-1:0] leftReadAddrDelay[LTDELAY-1:0];\n");
    }
    else if (toplat < leftlat) {
        fprintf(fp, "reg[TOPSIZEWIDTH-1:0] topReadAddrDelay[LTDELAY-1:0];\n");
    }
    fprintf(fp, "reg[TOPSIZEWIDTH-1:0] topWriteAddrDelay[MADELAY-1:0];\n");
    fprintf(fp, "reg topWriteEnDelay[MADELAY-1:0], topSourceSelDelay[CURLAT-1:0];\n");
    fprintf(fp, "reg[PEWIDTH-1:0] topWriteSelDelay[MADELAY-1:0];\n");
    fprintf(fp, "reg diagEnDelay[TOPLAT-1:0];\n");
    fprintf(fp, "reg MOEnDelay[MAXTOPLEFTLAT-1:0];\n");
    fprintf(fp, "reg [RAMSIZEWIDTH-1:0] waitCycles;\n");
    fprintf(fp, "\n");
    fprintf(fp, "// register store m, n and mdivk value\n");
    fprintf(fp, "always @ (posedge clk)\n");
    fprintf(fp, "begin\n");
    fprintf(fp, "	if (start_in == 1)\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		n <= n_in;\n");
    fprintf(fp, "		m <= m_in;\n");
    fprintf(fp, "		loop <= loop_in;\n");
    fprintf(fp, "		mode <= mode_in;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	if (mode[0] == 0 && m == loop)\n");
    fprintf(fp, "		stop <= loop;\n");
    fprintf(fp, "	else\n");
    fprintf(fp, "		stop <= loop+1;\n");
    fprintf(fp, "	stop2 <= loop;\n");
    fprintf(fp, "	startDelay[0] <= start_in;\n");
    for(i = 0; i < intdivlat; i++)
    {
        fprintf(fp, "	startDelay[%i] <= startDelay[%i];\n", i+1, i);
    }
    fprintf(fp, "	start <= startDelay[%i];\n", intdivlat);
    fprintf(fp, "	mdivk <= (m+NUMPE-1)/NUMPE;\n");
    fprintf(fp, "end\n\n");
    fprintf(fp, "// registers that store values that are used in FSM, dependent on i and/or j\n");
    fprintf(fp, "always @ (posedge clk)\n");
    fprintf(fp, "begin\n");
    fprintf(fp, "	if (start == 1)\n");
    fprintf(fp, "		topIdx <= 0; //offset1divk;\n");
    fprintf(fp, "	else if (currentState == INCRE_I && i1modk == NUMPE-1 && mode[0] == 0)\n");
    fprintf(fp, "		topIdx <= topIdx + 1;\n");
    fprintf(fp, "		\n");
    fprintf(fp, "	if (start == 1)\n");
    fprintf(fp, "		diagIdx <= 0; //offsetdivk;\n");
    fprintf(fp, "	else if (currentState == STORE_DIAG && mode == 1)\n");
    fprintf(fp, "		diagIdx <= BLOCKDIVK; //uppermdivk;\n");
    fprintf(fp, "	else if (currentState == INCRE_I)\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		if ((imodk == NUMPE-1 && mode == 0) || (i1modk == NUMPE-1 && mode == 1))\n");
    fprintf(fp, "			diagIdx <= diagIdx + BLOCKDIVK + 1;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			diagIdx <= diagIdx + BLOCKDIVK;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	\n");
    fprintf(fp, "	if (start == 1)\n");
    fprintf(fp, "		leftIdx <= 0;\n");
    fprintf(fp, "	else if (currentState == INCRE_I)\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		if (i1modk == NUMPE-1 && mode[0] == 0)\n");
    fprintf(fp, "			leftIdx <= leftIdx + BLOCKDIVK + 1;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			leftIdx <= leftIdx + BLOCKDIVK;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "\n");
    fprintf(fp, "	if (start == 1)\n");
    fprintf(fp, "		msIdx <= 0;\n");
    fprintf(fp, "	else if (currentState == UPDATE_J)\n");
    fprintf(fp,	"		if (mode[1] == 0)\n");
    fprintf(fp, "			msIdx <= leftIdx + BLOCKDIVK;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			msIdx <= topIdx;\n");
    fprintf(fp, "	else if (nextRowState == LOAD_ROW_INC_J)\n");
    fprintf(fp, "		msIdx <= msIdx + BLOCKDIVK;\n");
    fprintf(fp, "\n");
    fprintf(fp, "	if (start == 1)\n");
    fprintf(fp, "		imodk <= 0; //offsetmodk;\n");
    fprintf(fp, "	else if (currentState == INCRE_I)\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		if (imodk == NUMPE-1)\n");
    fprintf(fp, "			imodk <= 0;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			imodk <= imodk + 1;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	\n");
    fprintf(fp, "	if (start == 1)\n");
    fprintf(fp, "		i1modk <= 1; //offset1modk;\n");
    fprintf(fp, "	else if (currentState == INCRE_I)\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		if (i1modk == NUMPE-1)\n");
    fprintf(fp, "			i1modk <= 0;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			i1modk <= i1modk + 1;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	\n");
    fprintf(fp, "	if (start == 1)\n");
    fprintf(fp, "			nextTopIdx <= 0;\n");
    fprintf(fp, "	else if (currentState == INCRE_I)\n");
    fprintf(fp, "		if (mode[1] == 0)\n");
    fprintf(fp, "			nextTopIdx <= nextTopIdx + n + 1;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			nextTopIdx <= nextTopIdx + n;\n");
    fprintf(fp, " nextTopIdx2 <= nextTopIdx + n + 1;\n");
    fprintf(fp, "\n");
    fprintf(fp, "	if (start == 1)\n");
    fprintf(fp, "		curTopIdx <= 1;\n");
    fprintf(fp, "	else if (currentState == UPDATE_J)\n");
    fprintf(fp, "   if (mode[1] == 0)\n");
    fprintf(fp, "		  curTopIdx <= nextTopIdx+1;\n");
    fprintf(fp, "   else\n");
    fprintf(fp, "		  curTopIdx <= nextTopIdx;\n");
    fprintf(fp, "	else if (nextRowState == LOAD_ROW_INC_J)\n");
    fprintf(fp, "		curTopIdx <= curTopIdx + 1;\n");
    fprintf(fp, "	\n");
    fprintf(fp, "	if (start == 1)\n");
    fprintf(fp, "	   i1 <= 1;\n");
    fprintf(fp, "	else if (currentState == INCRE_I)\n");
    fprintf(fp, "	   i1 <= i1 + 1;\n");
    fprintf(fp, "\n");
    fprintf(fp, "	if (start == 1)\n");
    fprintf(fp, "		j <= 0;\n");
    fprintf(fp, "	else if (currentState == UPDATE_J)\n");
    fprintf(fp, "		if (mode[1] == 0)\n");
    fprintf(fp, "			j <= i1;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			j <= 0;\n");
    fprintf(fp, "	else if (currentRowState == LOAD_ROW_INC_J)\n");
    fprintf(fp, "		j <= j + 1;\n");
    fprintf(fp, "\n");
    fprintf(fp, "// compute cycles of delay in FSM\n");
    fprintf(fp, "	if (currentState == STORE_MO)\n");
    fprintf(fp, "		waitCycles <= MADELAY-1;\n");
    fprintf(fp, "	else if (currentState == INCRE_I)\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		if (i1 == stop-1)\n");
    fprintf(fp, "			if (mode[1] == 1)\n");
    fprintf(fp, "				waitCycles <= MADELAY-1 + MAXCURTOPLAT - 3;\n");
    fprintf(fp, "			else\n");
    fprintf(fp, "				waitCycles <= waitCycles + MAXCURLEFTLAT - 2;\n");
    fprintf(fp, "		else if (mode == 1 && waitCycles < MADELAY-1 - (MCDELAY-1) - 4)\n");
    fprintf(fp, "			waitCycles <= MADELAY-1 - (MCDELAY-1) - 4;\n");
    fprintf(fp, "		else if (mode == 2 && i1modk == NUMPE-1)\n");
    fprintf(fp, "			waitCycles <= MADELAY-1 + MAXCURTOPLAT - 3;\n");
    fprintf(fp, "		else if (mode == 0)\n");
    fprintf(fp, "			waitCycles <= waitCycles + MAXCURTOPLAT ;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	else if (waitCycles > 0)\n");
    fprintf(fp, "		waitCycles <= waitCycles - 1;\n");
    fprintf(fp, "\n");
    fprintf(fp, "end\n");
    fprintf(fp, "\n");
    fprintf(fp, "// determining next state of main FSM\n");
    fprintf(fp, "always @ (*)\n");
    fprintf(fp, "begin\n");
    fprintf(fp, "	case (currentState)\n");
    fprintf(fp, "	SETUP:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		if (start == 1)\n");
    fprintf(fp, "			nextState = START;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			nextState = SETUP;\n");
    fprintf(fp, "		updateCounter = 1;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	START:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp,	"		if (mode == 0)\n");
    fprintf(fp, "		begin\n");
    fprintf(fp, "			if (m == 1 && n == 1)\n");
    fprintf(fp, "				nextState = DONE;\n");
    fprintf(fp, "			else\n");
    fprintf(fp, "				nextState = FETCH_COL;\n");
    fprintf(fp, "		end\n");
    fprintf(fp, "		else if (mode == 1)\n");
    fprintf(fp,	"			nextState = STORE_DIAG;\n");
    fprintf(fp, "		else if (mode == 2)\n");
    fprintf(fp,	"			nextState = START_FETCH_ROW;\n");
    fprintf(fp, "		else\n");
    fprintf(fp,	"			nextState = UPDATE_J;\n");
    fprintf(fp, "		updateCounter = 1;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	START_FETCH_ROW:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		if (counter == CURLAT+TOPLAT-1)\n");
    fprintf(fp, "   begin\n");
    fprintf(fp, "		  if (mode == 0)\n");
    fprintf(fp, "			  nextState = STORE_DIAG;\n");
    fprintf(fp, "		  else\n");
    fprintf(fp, "			  nextState = UPDATE_J;\n");
    fprintf(fp, "	  end\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			nextState = START_FETCH_ROW;\n");
    fprintf(fp, "		updateCounter = 0;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	FETCH_COL:\n");
    fprintf(fp, "		if (counter >= mdivk-1)\n");
    fprintf(fp, "		begin\n");
    fprintf(fp, "			if (mode == 0 && counter < CURLAT)\n");
    fprintf(fp, "			begin\n");
    fprintf(fp, "				nextState = WAIT_COL;\n");
    fprintf(fp, "				updateCounter = 0;\n");
    fprintf(fp, "			end\n");
    fprintf(fp, "			else\n");
    fprintf(fp, "			begin\n");
    fprintf(fp, "				if (mode == 0)\n");
    fprintf(fp, "					nextState = START_FETCH_ROW;\n");
    fprintf(fp, "				else\n");
    fprintf(fp, "					nextState = FIND_REC;\n");
    fprintf(fp, "				updateCounter = 1;\n");
    fprintf(fp, "			end\n");
    fprintf(fp, "		end\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "		begin\n");
    fprintf(fp, "			nextState = FETCH_COL;\n");
    fprintf(fp, "			updateCounter = 0;\n");
    fprintf(fp, "		end\n");
    fprintf(fp, "	WAIT_COL:\n");
    fprintf(fp, "		if (counter >= CURLAT)\n");
    fprintf(fp, "		begin\n");
    fprintf(fp, "			if (mode == 0)\n");
    fprintf(fp, "				nextState = START_FETCH_ROW;\n");
    fprintf(fp, "			else\n");
    fprintf(fp, "				nextState = FIND_REC;\n");
    fprintf(fp, "			updateCounter = 1;\n");
    fprintf(fp, "		end\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "		begin\n");
    fprintf(fp, "			nextState = WAIT_COL;\n");
    fprintf(fp, "			updateCounter = 0;\n");
    fprintf(fp, "		end\n");
    fprintf(fp, "	STORE_DIAG:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		if (mode == 0)\n");
    fprintf(fp, "			nextState =  FIND_REC;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			nextState =  FETCH_COL;\n");
    fprintf(fp, "		updateCounter = 1;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	FIND_REC:\n");
    fprintf(fp, "		if (divCounter == DIVLAT)\n");
    fprintf(fp, "		begin\n");
    fprintf(fp, "			if (mode == 0)\n");
    fprintf(fp, "				nextState = MULT_COL;\n");
    fprintf(fp, "			else\n");
    fprintf(fp, "				nextState = STORE_DIAG2;\n");
    fprintf(fp, "			updateCounter = 1;\n");
    fprintf(fp, "		end\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "		begin\n");
    fprintf(fp, "			nextState = FIND_REC;\n");
    fprintf(fp, "			updateCounter = 0;\n");
    fprintf(fp, "		end\n");
    fprintf(fp, "	STORE_DIAG2:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		nextState =  MULT_COL;\n");
    fprintf(fp, "		updateCounter = 1;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	MULT_COL:\n");
    fprintf(fp, "		if (topIdxCounter == mdivk-1)\n");
    fprintf(fp, "		begin\n");
    fprintf(fp, "			nextState = UPDATE_J;\n");
    fprintf(fp, "			updateCounter = 0;\n");
    fprintf(fp, "		end\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "		begin\n");
    fprintf(fp, "			nextState = MULT_COL;\n");
    fprintf(fp, "			updateCounter = 0;\n");
    fprintf(fp, "		end\n");
    fprintf(fp, "	UPDATE_J:\n");
    fprintf(fp, "		if ((mode[1] == 1 || counter >= MCDELAY-1) && doneFetchRow == 1)\n");
    fprintf(fp, "		begin\n");
    fprintf(fp, "			nextState = STORE_MO;\n");
    fprintf(fp, "			updateCounter = 1;\n");
    fprintf(fp, "		end\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "		begin\n");
    fprintf(fp, "			nextState = UPDATE_J;\n");
    fprintf(fp, "			updateCounter = 0;\n");
    fprintf(fp, "		end\n");
    fprintf(fp, "	STORE_MO:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		if (j == stop2)\n");
    fprintf(fp, "		begin\n");
    fprintf(fp, "			if (counter == mdivk-1+MAXCURLEFTLAT-2)\n");
    fprintf(fp, "				nextState = DONE;\n");
    fprintf(fp, "			else\n");
    fprintf(fp, "				nextState = STORE_MO;\n");
    fprintf(fp, "			updateCounter = 0;\n");
    fprintf(fp, "		end\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "		begin\n");
    fprintf(fp, "			nextState =  MULT_SUB;\n");
    fprintf(fp, "			updateCounter = 1;\n");
    fprintf(fp, "		end\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	MULT_SUB:\n");
    fprintf(fp, "		if (topIdxCounter == mdivk-1)\n");
    fprintf(fp, "		begin\n");
    fprintf(fp, "			if (j == n-1)\n");
    fprintf(fp, "				nextState = INCRE_I;\n");
    fprintf(fp, "			else\n");
    fprintf(fp, "				nextState = MULT_SUB;\n");
    fprintf(fp, "			updateCounter = 1;\n");
    fprintf(fp, "		end\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "		begin\n");
    fprintf(fp, "			nextState = MULT_SUB;\n");
    fprintf(fp, "			updateCounter = 0;\n");
    fprintf(fp, "		end\n");
    fprintf(fp, "	INCRE_I:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		nextState = WAIT;\n");
    fprintf(fp, "		updateCounter = 1;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	WAIT:\n");
    fprintf(fp, "		if (waitCycles == 0)\n");
    fprintf(fp, "		begin\n");
    fprintf(fp, "			if (i1 == stop)\n");
    fprintf(fp, "				nextState = DONE;\n");
    fprintf(fp, "			else if (mode == 0)\n");
    fprintf(fp, "				nextState = STORE_DIAG;\n");
    fprintf(fp, "			else if (mode == 1)\n");
    fprintf(fp, "				nextState = FIND_REC;\n");
    fprintf(fp, "			else\n");
    fprintf(fp, "				nextState = UPDATE_J;\n");
    fprintf(fp, "			updateCounter = 1;\n");
    fprintf(fp, "		end\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "		begin\n");
    fprintf(fp, "			nextState = WAIT;\n");
    fprintf(fp, "			updateCounter = 0;\n");
    fprintf(fp, "		end\n");
    fprintf(fp, "	DONE:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		nextState = DONE;\n");
    fprintf(fp, "		updateCounter = 0;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	default:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		nextState = SETUP;\n");
    fprintf(fp, "		updateCounter = 1;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	endcase\n");
    fprintf(fp, "end\n");
    fprintf(fp, "\n");
    fprintf(fp, "always @ (*)\n");
    fprintf(fp, "begin\n");
    fprintf(fp, "	if (currentRowState == DONE_FETCH_ROW)\n");
    fprintf(fp, "		doneFetchRow = 1;\n");
    fprintf(fp, "	else\n");
    fprintf(fp, "		doneFetchRow = 0;\n");
    fprintf(fp, "		if((nextState == START_FETCH_ROW && currentState != START_FETCH_ROW && i1 == 1))\n");
    fprintf(fp, "		startFetchRow = 1;\n");
    fprintf(fp, "	else\n");
    fprintf(fp, "		startFetchRow = 0;\n");
    fprintf(fp, "	if (currentState == MULT_SUB && topIdxCounter+2 == mdivk)\n");
    fprintf(fp, "		loadRow = 1;\n");
    fprintf(fp, "	else\n");
    fprintf(fp, "		loadRow = 0;\n");
    fprintf(fp, "	writeRow = (msIdxCounter == readRowCounter)&&(currentState==MULT_SUB)&&(j!=n)&&(mode[0] == 0);\n");
    fprintf(fp, "end\n");
    fprintf(fp, "\n");
    fprintf(fp, "// second FSM that controls the control signals to temp_top block\n");
    fprintf(fp, "always @ (*)\n");
    fprintf(fp, "begin\n");
    fprintf(fp, "	case (currentRowState)\n");
    fprintf(fp, "	FETCH_ROW:\n");
    fprintf(fp, "		if (nextTopIdxCounter == n-1)\n");
    fprintf(fp, "			nextRowState = DONE_FETCH_ROW;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			nextRowState = FETCH_ROW;\n");
    fprintf(fp, "	DONE_FETCH_ROW:\n");
    fprintf(fp, "		if (startFetchRow == 1)\n");
    fprintf(fp, "			nextRowState = FETCH_ROW;\n");
    fprintf(fp, "		else if (loadRow == 1 || (topIdx+1 == mdivk && nextState == MULT_SUB))\n");
    fprintf(fp, "			nextRowState = LOAD_ROW_INC_J;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			nextRowState = DONE_FETCH_ROW;\n");
    fprintf(fp, "	LOAD_ROW_INC_J:\n");
    fprintf(fp, "		if (topIdx+1 == mdivk && nextState == MULT_SUB)\n");
    fprintf(fp, "			nextRowState = LOAD_ROW_INC_J;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			nextRowState = DONE_FETCH_ROW;\n");
    fprintf(fp, "	default:\n");
    fprintf(fp, "		nextRowState = DONE_FETCH_ROW;\n");
    fprintf(fp, "	endcase\n");
    fprintf(fp, "end\n");
    fprintf(fp, "\n");
    fprintf(fp, "// address counters\n");
    fprintf(fp, "always @ (posedge clk)\n");
    fprintf(fp, "begin\n");
    fprintf(fp, "	if (updateCounter == 1 || currentRowState == LOAD_ROW_INC_J)\n");
    fprintf(fp, "		topIdxCounter <= topIdx;\n");
    fprintf(fp, "	else\n");
    fprintf(fp, "		topIdxCounter <= topIdxCounter + 1;\n");
    fprintf(fp, "\n");
    fprintf(fp, "	if (updateCounter == 1)\n");
    fprintf(fp, "		diagIdxCounter <= diagIdx;\n");
    fprintf(fp, "	else\n");
    fprintf(fp, "		diagIdxCounter <= diagIdxCounter + 1;\n");
    fprintf(fp, "\n");
    fprintf(fp, "	if (updateCounter == 1 || currentRowState == LOAD_ROW_INC_J)\n");
    fprintf(fp, "		msIdxCounter <= msIdx;\n");
    fprintf(fp, "	else\n");
    fprintf(fp, "		msIdxCounter <= msIdxCounter + 1;\n");
    fprintf(fp, "\n");
    fprintf(fp, "	if (updateCounter == 1 || currentRowState == LOAD_ROW_INC_J)\n");
    fprintf(fp, "		leftIdxCounter <= leftIdx;\n");
    fprintf(fp, "	else\n");
    fprintf(fp, "		leftIdxCounter <= leftIdxCounter + 1;\n");
    fprintf(fp, "	\n");
    fprintf(fp, "	if (currentState == FETCH_COL || currentState == STORE_MO)\n");
    fprintf(fp, "		topWriteCounter <= i1;\n");
    fprintf(fp, "	else if (writeRow == 1 || currentRowState == FETCH_ROW)\n");
    fprintf(fp, "		topWriteCounter <= topWriteCounter + 1;\n");
    fprintf(fp, "\n");
    fprintf(fp, "	if (currentState == START)\n");
    fprintf(fp, "		nextTopIdxCounter <= nextTopIdx;\n");
    fprintf(fp, " else if (currentState == STORE_MO)\n");
    fprintf(fp, "		if (mode[1] == 0)\n");
    fprintf(fp, "			nextTopIdxCounter <= nextTopIdx + n + 1;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			nextTopIdxCounter <= nextTopIdx + n;\n");
    fprintf(fp, "	else if (writeRow == 1 || currentRowState == FETCH_ROW)\n");
    fprintf(fp, "		nextTopIdxCounter <= nextTopIdxCounter + 1;\n");
    fprintf(fp, "\n");
    fprintf(fp, "	if (currentState == START)\n");
    fprintf(fp, "			readRowCounter <= 0; //offsetdivk;\n");
    fprintf(fp, "	else if (currentState == STORE_MO)\n");
    fprintf(fp, "		if (mode[1] == 0)\n");
    fprintf(fp, "			readRowCounter <= leftIdx + BLOCKDIVK;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			readRowCounter <= topIdx;\n");
    fprintf(fp, "	else if (writeRow == 1 || currentRowState == FETCH_ROW)\n");
    fprintf(fp, "	   readRowCounter <= readRowCounter + BLOCKDIVK;\n");
    fprintf(fp, "\n");
    fprintf(fp, "	if (updateCounter == 1)\n");
    fprintf(fp, "		counter <= 0;\n");
    fprintf(fp, "	else\n");
    fprintf(fp, "		counter <= counter + 1;\n");
    fprintf(fp, "\n");
    fprintf(fp, "	if (currentState == STORE_DIAG || currentState == STORE_DIAG2)\n");
    fprintf(fp, "		divCounter <= 0;\n");
    fprintf(fp, "	else if (divCounter < DIVLAT)\n");
    fprintf(fp, "		divCounter <= divCounter + 1;\n");
    fprintf(fp, "\n");
    if (precision == 64)
    {
        fprintf(fp, "	i1modkByteEn <= ~(%i\'h0) >> (i1modk*8);\n", ramwidth/8);
    }
    else
    {
        fprintf(fp, "	i1modkByteEn <= ~(%i\'h0) >> (i1modk*4);\n", ramwidth/8);
    }
    fprintf(fp, "end\n");
    fprintf(fp, "\n");
    fprintf(fp, "// compute Byte Enable\n");
    fprintf(fp, "always @ (posedge clk)\n");
    fprintf(fp, "begin\n");
    fprintf(fp, "	if ((nextState == MULT_COL && currentState != MULT_COL) || (currentState == STORE_MO) || currentRowState == LOAD_ROW_INC_J)\n");
    fprintf(fp, "		byteEn <= i1modkByteEn;\n");
    fprintf(fp, "	else\n");
    fprintf(fp, "		byteEn <= ~0;\n");
    fprintf(fp, "end\n");
    fprintf(fp, "\n");
    fprintf(fp, "// update FSM state register\n");
    fprintf(fp, "always @ (posedge clk)\n");
    fprintf(fp, "begin\n");
    fprintf(fp, "	if (start_in == 1)\n");
    fprintf(fp, "		currentState <= SETUP;\n");
    fprintf(fp, "	else\n");
    fprintf(fp, "		currentState <= nextState;\n");
    fprintf(fp, "	if (start == 1)\n");
    fprintf(fp, "		currentRowState <= DONE_FETCH_ROW;\n");
    fprintf(fp, "	else\n");
    fprintf(fp, "		currentRowState <= nextRowState;\n");
    fprintf(fp, "end\n");
    fprintf(fp, "\n");
    fprintf(fp, "// delay register for control signals\n");
    fprintf(fp, "// control signals are delayed to match latency of operations and/or memory access\n");
    fprintf(fp, "always @ (posedge clk)\n");
    fprintf(fp, "begin\n");

    for (i = 0; i < mdelay-1; i++) {
        fprintf(fp, "	curReadAddrDelay[%i] <= curReadAddrDelay[%i];\n", i, i+1);
    }
    fprintf(fp, "	curReadAddrDelay[%i] <= msIdxCounter;\n", i);
    fprintf(fp, "	\n");

    for (i = 0; i < curlat-1; i++) {
        fprintf(fp, "	curWriteAddrDelay[%i] <= curWriteAddrDelay[%i];\n", i, i+1);
    }
    fprintf(fp, "	if (currentState == FETCH_COL)\n");
    fprintf(fp, "		curWriteAddrDelay[%i] <= diagIdxCounter;\n", i);
    fprintf(fp, "	else\n");
    fprintf(fp, "		curWriteAddrDelay[%i] <= curWriteAddrDelay[%i];\n", i, i+1);
    for (i = curlat; i < mcdelay-1; i++) {
        fprintf(fp, "	curWriteAddrDelay[%i] <= curWriteAddrDelay[%i];\n", i, i+1);
    }
    fprintf(fp, "	if (currentState == MULT_COL)\n");
    fprintf(fp, "		curWriteAddrDelay[%i] <= leftIdxCounter;\n", i);
    fprintf(fp, "	else\n");
    fprintf(fp, "		curWriteAddrDelay[%i] <= curWriteAddrDelay[%i];\n", i, i+1);
    for (i = mcdelay; i < madelay-1; i++) {
        fprintf(fp, "	curWriteAddrDelay[%i] <= curWriteAddrDelay[%i];\n", i, i+1);
    }
    fprintf(fp, "	curWriteAddrDelay[%i] <= msIdxCounter;\n", i);
    fprintf(fp, "	\n");

    for (i = 0; i < curlat-1; i++) {
        fprintf(fp, "	writeByteEnDelay[%i] <= writeByteEnDelay[%i];\n", i, i+1);
    }
    fprintf(fp, "	if (mode[0] == 1)\n");
    fprintf(fp, "		writeByteEnDelay[%i] <= ~0;\n", i);
    fprintf(fp, "	else if (currentState == FETCH_COL)\n");
    fprintf(fp, "		writeByteEnDelay[%i] <= byteEn;\n", i);
    fprintf(fp, "	else\n");
    fprintf(fp, "		writeByteEnDelay[%i] <= writeByteEnDelay[%i];\n", i, i+1);
    for (i = curlat; i < mcdelay-1; i++) {
        fprintf(fp, "	writeByteEnDelay[%i] <= writeByteEnDelay[%i];\n", i, i+1);
    }
    fprintf(fp, "	if (currentState == MULT_COL)\n");
    fprintf(fp, "		writeByteEnDelay[%i] <= byteEn;\n", i);
    fprintf(fp, "	else\n");
    fprintf(fp, "		writeByteEnDelay[%i] <= writeByteEnDelay[%i];\n", i, i+1);
    for (i = mcdelay; i < madelay-1; i++) {
        fprintf(fp, "	writeByteEnDelay[%i] <= writeByteEnDelay[%i];\n", i, i+1);
    }
    fprintf(fp, "	writeByteEnDelay[%i] <= byteEn;\n", i);
    fprintf(fp, "	\n");

    for (i = 0; i < mcdelay-1; i++) {
        fprintf(fp, "	curWriteSelDelay[%i] <= curWriteSelDelay[%i];\n", i, i+1);
    }
    fprintf(fp, "	if (currentState == MULT_COL)\n");
    fprintf(fp, "		curWriteSelDelay[%i] <= 0;\n", i);
    fprintf(fp, "	else\n");
    fprintf(fp, "		curWriteSelDelay[%i] <= 1;\n", i);
    fprintf(fp, "\n");

    for (i = 0; i < mcdelay-1; i++) {
        fprintf(fp, "	curWriteEnDelay[%i] <= curWriteEnDelay[%i];\n", i, i+1);
    }
    fprintf(fp, "	if (currentState == MULT_COL)\n");
    fprintf(fp, "		curWriteEnDelay[%i] <= 1;\n", i);
    fprintf(fp, "	else\n");
    fprintf(fp, "		curWriteEnDelay[%i] <= curWriteEnDelay[%i];\n", i, i+1);
    for (i = mcdelay; i < madelay-1; i++) {
        fprintf(fp, "	curWriteEnDelay[%i] <= curWriteEnDelay[%i];\n", i, i+1);
    }
    fprintf(fp, "	if (currentState == MULT_SUB)\n");
    fprintf(fp, "		curWriteEnDelay[%i] <= 1;\n", i);
    fprintf(fp, "	else\n");
    fprintf(fp, "		curWriteEnDelay[%i] <= 0;\n", i);
    fprintf(fp, "\n");

    for (i = 0; i < curlat-1; i++) {
        fprintf(fp, "	leftWriteSelDelay[%i] <= leftWriteSelDelay[%i];\n", i, i+1);
    }
    fprintf(fp, "	if (currentState == FETCH_COL)\n");
    fprintf(fp, "		leftWriteSelDelay[%i] <= 0;\n", i);
    fprintf(fp, "	else\n");
    fprintf(fp, "		leftWriteSelDelay[%i] <= 1;\n", i);
    fprintf(fp, "\n");

    for (i = 0; i < curlat-1; i++) {
        fprintf(fp, "	leftWriteEnDelay[%i] <= leftWriteEnDelay[%i];\n", i, i+1);
    }
    fprintf(fp, "	if (currentState == FETCH_COL)\n");
    fprintf(fp, "		leftWriteEnDelay[%i] <= 1;\n", i);
    fprintf(fp, "	else\n");
    fprintf(fp, "		leftWriteEnDelay[%i] <= leftWriteEnDelay[%i];\n", i, i+1);
    for (i = curlat; i < mcdelay-1; i++) {
        fprintf(fp, "	leftWriteEnDelay[%i] <= leftWriteEnDelay[%i];\n", i, i+1);
    }
    fprintf(fp, "	if (currentState == MULT_COL)\n");
    fprintf(fp, "		leftWriteEnDelay[%i] <= 1;\n", i);
    fprintf(fp, "	else\n");
    fprintf(fp, "		leftWriteEnDelay[%i] <= leftWriteEnDelay[%i];\n", i, i+1);
    for (i = mcdelay; i < madelay-1; i++) {
        fprintf(fp, "	leftWriteEnDelay[%i] <= leftWriteEnDelay[%i];\n", i, i+1);
    }
    fprintf(fp, "	if (currentState == MULT_SUB && (mode == 0 || (mode == 1 && j == i1)))\n");
    fprintf(fp, "		leftWriteEnDelay[%i] <= 1;\n", i);
    fprintf(fp, "	else\n");
    fprintf(fp, "		leftWriteEnDelay[%i] <= 0;\n", i);
    fprintf(fp, "\n");

    for (i = 0; i < curlat-1; i++) {
        fprintf(fp, "	topWriteAddrDelay[%i] <= topWriteAddrDelay[%i];\n", i, i+1);
    }
    fprintf(fp, "	if (currentRowState == FETCH_ROW)\n");
    fprintf(fp, "		topWriteAddrDelay[%i] <= nextTopIdxCounter;\n", i);
    fprintf(fp, "	else\n");
    fprintf(fp, "		topWriteAddrDelay[%i] <=  topWriteAddrDelay[%i];\n", i, i+1);
    for (i = curlat; i < madelay-1; i++) {
        fprintf(fp, "	topWriteAddrDelay[%i] <= topWriteAddrDelay[%i];\n", i, i+1);
    }
    fprintf(fp, "		topWriteAddrDelay[%i] <= nextTopIdxCounter;\n", i);
    fprintf(fp, "\n");

    for (i = 0; i < curlat-1; i++) {
        fprintf(fp, "	topWriteEnDelay[%i] <= topWriteEnDelay[%i];\n", i, i+1);
    }
    fprintf(fp, "	if (currentRowState == FETCH_ROW)\n");
    fprintf(fp, "		topWriteEnDelay[%i] <= 1;\n", i);
    fprintf(fp, "	else\n");
    fprintf(fp, "		topWriteEnDelay[%i] <=  topWriteEnDelay[%i];\n", i, i+1);
    for (i = curlat; i < madelay-1; i++) {
        fprintf(fp, "	topWriteEnDelay[%i] <= topWriteEnDelay[%i];\n", i, i+1);
    }
    fprintf(fp, "	topWriteEnDelay[%i] <= writeRow;\n", i);
    fprintf(fp, "\n");

    for (i = 0; i < curlat-1; i++) {
        fprintf(fp, "	topWriteSelDelay[%i] <= topWriteSelDelay[%i];\n", i, i+1);
    }
    fprintf(fp, "	if (currentRowState == FETCH_ROW || currentState == UPDATE_J && i1 == 1)\n");
    fprintf(fp, "		topWriteSelDelay[%i] <= imodk;\n", i);
    fprintf(fp, "	else\n");
    fprintf(fp, "		topWriteSelDelay[%i] <=  topWriteSelDelay[%i];\n", i, i+1);
    for (i = curlat; i < madelay-1; i++) {
        fprintf(fp, "	topWriteSelDelay[%i] <= topWriteSelDelay[%i];\n", i, i+1);
    }
    fprintf(fp,	"	topWriteSelDelay[%i] <= i1modk;\n", i);
    fprintf(fp, "\n");

    for (i = 0; i < curlat-1; i++) {
        fprintf(fp, "	topSourceSelDelay[%i] <= topSourceSelDelay[%i];\n", i, i+1);
    }
    fprintf(fp, "	if (start == 1)\n");
    fprintf(fp, "		topSourceSelDelay[%i] <= 0;\n", i);
    fprintf(fp, "	else if (currentState == STORE_MO)\n");
    fprintf(fp,	"		topSourceSelDelay[%i] <= 1;\n", i);
    fprintf(fp, "\n");

    if (leftlat > toplat) {
        for (i = 0; i < ltdelay-1; i++) {
            fprintf(fp, "	topReadAddrDelay[%i] <= topReadAddrDelay[%i];\n", i, i+1);
        }
        fprintf(fp, "	topReadAddrDelay[%i] <= curTopIdx;\n", i);
        fprintf(fp, "\n");
    }

    if (toplat > leftlat) {
        for (i = 0; i < ltdelay-1; i++) {
            fprintf(fp, "	leftReadAddrDelay[%i] <= leftReadAddrDelay[%i];\n", i, i+1);
        }
        fprintf(fp, "	leftReadAddrDelay[%i] <= leftIdxCounter;\n", i);
        fprintf(fp, "\n");
    }
    fprintf(fp, "\n");

    for (i = 0; i < toplat-1; i++) {
        fprintf(fp, "	diagEnDelay[%i] <= diagEnDelay[%i];\n", i, i+1);
    }
    fprintf(fp, "	diagEnDelay[%i] <= (currentState == STORE_DIAG || currentState == STORE_DIAG2);\n", i);
    fprintf(fp, "\n");

    for (i = 0; i < maxtopleftlat-1; i++) {
        fprintf(fp, "	MOEnDelay[%i] <= MOEnDelay[%i];\n", i, i+1);
    }
    fprintf(fp, "	if (currentState == STORE_MO || currentRowState == LOAD_ROW_INC_J)\n");
    fprintf(fp, "		MOEnDelay[%i] <= 1;\n", i);
    fprintf(fp, "	else\n");
    fprintf(fp, "		MOEnDelay[%i] <= 0;\n", i);
    
    fprintf(fp, "end\n");
    fprintf(fp, "\n");

    fprintf(fp, "// output contorl signals\n");
    fprintf(fp, "always @ (*)\n");
    fprintf(fp, "begin\n");
    fprintf(fp, "	if (currentState == FETCH_COL)\n");
    fprintf(fp, "		curReadAddr <= diagIdxCounter;\n");
    fprintf(fp, "	else if (currentRowState == FETCH_ROW)\n");
    fprintf(fp, "	   curReadAddr <= readRowCounter;\n");
    fprintf(fp, "	else\n");
    fprintf(fp, "		curReadAddr <= curReadAddrDelay[0];\n");
    fprintf(fp, "	curWriteAddr <= curWriteAddrDelay[0];\n");
    fprintf(fp, "	curWriteByteEn <= writeByteEnDelay[0];\n");
    fprintf(fp, "	curWriteSel <= curWriteSelDelay[0];\n");
    fprintf(fp, "	curWriteEn <= curWriteEnDelay[0];\n");
    fprintf(fp, "\n");
    if (toplat > leftlat) {
        fprintf(fp, "	if (currentState == MULT_COL)\n");
        fprintf(fp, "		leftReadAddr <= leftIdxCounter;\n");
        fprintf(fp, "	else\n");
        fprintf(fp, "		leftReadAddr <= leftReadAddrDelay[0];\n");
    }
    else {
        fprintf(fp, "	leftReadAddr <= leftIdxCounter;\n");
    }
    fprintf(fp, "	leftWriteAddr <= curWriteAddrDelay[0];\n");
    fprintf(fp, "	leftWriteByteEn <= writeByteEnDelay[0];\n");
    fprintf(fp, "	leftWriteSel <= leftWriteSelDelay[0];\n");
    fprintf(fp, "	leftWriteEn <= leftWriteEnDelay[0];\n");
    fprintf(fp, "\n");

    fprintf(fp, "	if (currentState == STORE_DIAG)\n");
    fprintf(fp, "		topReadAddr <= nextTopIdx;\n");
    fprintf(fp, "else if (currentState == STORE_DIAG2)\n");
    fprintf(fp, "   topReadAddr <= nextTopIdx2;\n");
    fprintf(fp, "	else\n");
    if (leftlat > toplat) {
        fprintf(fp, "	topReadAddr <= topReadAddrDelay[0];\n");
    }
    else {
        fprintf(fp, "	topReadAddr <= curTopIdx;\n");
    }
    fprintf(fp, "	topWriteAddr <= topWriteAddrDelay[0];\n");
    fprintf(fp, "	topWriteEn <= topWriteEnDelay[0];\n");
    fprintf(fp, "	topWriteSel <= topWriteSelDelay[0];\n");
    fprintf(fp, "	topSourceSel <= topSourceSelDelay[0];\n");
    fprintf(fp, "\n");
    fprintf(fp, "	MOSel <= ~(currentState == FIND_REC);\n");
    fprintf(fp, "if (currentState == FIND_REC)\n");
    fprintf(fp, "		MOEn <= 1;\n");
    fprintf(fp, "	else\n");
    fprintf(fp, "		MOEn <= MOEnDelay[0];\n");
    fprintf(fp, "\n");

    fprintf(fp, "	diagEn <= diagEnDelay[0];\n");
    fprintf(fp, "\n");
    fprintf(fp, "	if (currentState == DONE)\n");
    fprintf(fp, "		done <= 1;\n");
    fprintf(fp, "	else\n");
    fprintf(fp, "		done <= 0;\n");
    fprintf(fp, "end\n");
    fprintf(fp, "\n");
    fprintf(fp, "endmodule\n");
    fclose(fp);
}
