// generate LU module
// by Wei Zhang
//

#include <stdio.h>
#include <math.h>
#define intlog2(x) (int)ceil(log(x)/log(2))

void genBlockLU(const char *divname, int numPE, int precision, int mwidth, int ramwidth, int ramsizewidth, int topsizewidth,
    int meminputdelay, int memoutputdelay, int topinputdelay, int topoutputdelay)
{
    FILE *fp;
    int PEwidth;
    int i, upperIdx, lowerIdx;


    PEwidth = intlog2(numPE);

    fp = fopen("LU.v", "w");

    fprintf(fp, "//auto-generated LU.v\n");
    fprintf(fp, "//datapath for computating LU factorization\n");
    fprintf(fp, "//by Wei Zhang\n\n");
    fprintf(fp, "module LU (clk, start, m, n, loop, mode, done, \n");
    fprintf(fp, "			curReadAddrMem, curReadDataMem, curWriteByteEnMem, curWriteDataMem, curWriteAddrMem, curWriteEnMem, curMemSel,\n");
    fprintf(fp, "			leftWriteByteEnMem, leftWriteDataMem, leftWriteAddrMem, leftWriteEnMem, leftMemSel);\n");
    fprintf(fp, "\n\n");

    fprintf(fp, "parameter PRECISION = %i, NUMPE = %i, PEWIDTH = %i, BLOCKWIDTH = %i;\n", precision, numPE, PEwidth, mwidth);
    fprintf(fp, "parameter RAMWIDTH = %i, RAMNUMBYTES = %i, RAMSIZEWIDTH = %i, TOPSIZEWIDTH = %i;\n", ramwidth, ramwidth/8, ramsizewidth, topsizewidth);
    fprintf(fp, "parameter TOPINPUTDELAY = %i, TOPOUTPUTDELAY = %i;\n", topinputdelay, topoutputdelay);
    fprintf(fp, "parameter MEMINPUTDELAY = %i, MEMOUTPUTDELAY = %i;\n", meminputdelay, memoutputdelay);
    fprintf(fp, "\n");
 
    fprintf(fp, "input clk, start;\n");
    fprintf(fp, "input[BLOCKWIDTH-1:0] m, n, loop;\n");
    fprintf(fp, "input[1:0] mode;\n");
    fprintf(fp, "output done;\n"); 
    fprintf(fp, "wire[RAMWIDTH-1:0] curWriteData0, curWriteData1;\n");
    fprintf(fp, "wire[RAMSIZEWIDTH-1:0] curWriteAddr0, curReadAddr0, curWriteAddr1, curReadAddr1;\n");
    fprintf(fp, "wire[RAMWIDTH-1:0] curReadData0, curReadData1;\n");
    fprintf(fp, "wire[RAMNUMBYTES-1:0] curWriteByteEn0, curWriteByteEn1;\n");
    fprintf(fp, "wire curWriteEn0, curWriteEn1;\n\n");

    fprintf(fp, "input[RAMWIDTH-1:0] curWriteDataMem;\n");
    fprintf(fp, "output[RAMWIDTH-1:0] curReadDataMem;\n");
    fprintf(fp, "input[RAMSIZEWIDTH-1:0] curWriteAddrMem, curReadAddrMem;\n");
    fprintf(fp, "input[RAMNUMBYTES-1:0] curWriteByteEnMem;\n");
    fprintf(fp, "input curWriteEnMem;\n");
    fprintf(fp, "input[RAMWIDTH-1:0] leftWriteDataMem;\n");
    fprintf(fp, "input[RAMSIZEWIDTH-1:0] leftWriteAddrMem;\n");
    fprintf(fp, "input[RAMNUMBYTES-1:0] leftWriteByteEnMem;\n");
    fprintf(fp, "input leftWriteEnMem;\n");
    fprintf(fp, "input leftMemSel, curMemSel;\n\n");

    fprintf(fp, "wire[RAMWIDTH-1:0] curReadDataLU, curReadDataMem;\n");
    fprintf(fp, "wire[RAMWIDTH-1:0] curWriteDataLU, curWriteDataMem;\n");
    fprintf(fp, "wire[RAMSIZEWIDTH-1:0] curWriteAddrLU, curWriteAddrMem, curReadAddrLU, curReadAddrMem;\n");
    fprintf(fp, "wire[RAMNUMBYTES-1:0] curWriteByteEnLU, curWriteByteEnMem;\n");
    fprintf(fp, "wire curWriteEnLU, curWriteEnMem;\n\n");

    if (memoutputdelay > 0) {
        fprintf(fp, "reg[RAMWIDTH-1:0] curReadData0Reg[MEMOUTPUTDELAY-1:0], curReadData1Reg[MEMOUTPUTDELAY-1:0];\n");
        fprintf(fp, "reg[RAMWIDTH-1:0] leftReadData0Reg[MEMOUTPUTDELAY-1:0], leftReadData1Reg[MEMOUTPUTDELAY-1:0];\n");
    }

    if (meminputdelay > 0)
    {
        fprintf(fp, "reg[RAMWIDTH-1:0] curWriteData0Reg[MEMINPUTDELAY-1:0], curWriteData1Reg[MEMINPUTDELAY-1:0];\n");
        fprintf(fp, "reg[RAMSIZEWIDTH-1:0] curWriteAddr0Reg[MEMINPUTDELAY-1:0], curReadAddr0Reg[MEMINPUTDELAY-1:0];\n");
        fprintf(fp, "reg[RAMSIZEWIDTH-1:0] curWriteAddr1Reg[MEMINPUTDELAY-1:0], curReadAddr1Reg[MEMINPUTDELAY-1:0];\n");
        fprintf(fp, "reg[RAMNUMBYTES-1:0] curWriteByteEn0Reg[MEMINPUTDELAY-1:0], curWriteByteEn1Reg[MEMINPUTDELAY-1:0];\n");
        fprintf(fp, "reg curWriteEn0Reg[MEMINPUTDELAY-1:0], curWriteEn1Reg[MEMINPUTDELAY-1:0];\n");
        fprintf(fp, "reg[RAMWIDTH-1:0] leftWriteData0Reg[MEMINPUTDELAY-1:0], leftWriteData1Reg[MEMINPUTDELAY-1:0];\n");
        fprintf(fp, "reg[RAMSIZEWIDTH-1:0] leftWriteAddr0Reg[MEMINPUTDELAY-1:0], leftReadAddr0Reg[MEMINPUTDELAY-1:0];\n");
        fprintf(fp, "reg[RAMSIZEWIDTH-1:0] leftWriteAddr1Reg[MEMINPUTDELAY-1:0], leftReadAddr1Reg[MEMINPUTDELAY-1:0];\n");
        fprintf(fp, "reg[RAMNUMBYTES-1:0] leftWriteByteEn0Reg[MEMINPUTDELAY-1:0], leftWriteByteEn1Reg[MEMINPUTDELAY-1:0];\n");
        fprintf(fp, "reg leftWriteEn0Reg[MEMINPUTDELAY-1:0], leftWriteEn1Reg[MEMINPUTDELAY-1:0];\n");
    }
    fprintf(fp, "\n");

    fprintf(fp, "reg[PRECISION-1:0] multOperand;\n");
    fprintf(fp, "reg[PRECISION-1:0] diag;\n");
    fprintf(fp, "wire[PRECISION-1:0] recResult;\n");
    fprintf(fp, "wire[PRECISION-1:0] multA[NUMPE-1:0], multResult[NUMPE-1:0];\n");
    fprintf(fp, "wire[PRECISION-1:0] addA[NUMPE-1:0], addResult[NUMPE-1:0];\n\n");

    fprintf(fp, "wire[RAMWIDTH-1:0] leftReadData0, leftReadData1, leftWriteData0, leftWriteData1;\n");
    fprintf(fp, "wire[RAMSIZEWIDTH-1:0] leftWriteAddr0, leftWriteAddr1, leftReadAddr0, leftReadAddr1;\n");
    fprintf(fp, "wire[RAMNUMBYTES-1:0] leftWriteByteEn0, leftWriteByteEn1;\n");
    fprintf(fp, "wire leftWriteEn0, leftWriteEn1;\n");
    fprintf(fp, "wire[RAMWIDTH-1:0] leftReadDataLU, leftWriteDataLU, leftWriteDataMem;\n");
    fprintf(fp, "wire[RAMSIZEWIDTH-1:0] leftWriteAddrLU, leftWriteAddrMem, leftReadAddrLU;\n");
    fprintf(fp, "wire[RAMNUMBYTES-1:0] leftWriteByteEnLU, leftWriteByteEnMem;\n");
    fprintf(fp, "wire leftWriteEnLU, leftWriteEnMem;\n\n");

    fprintf(fp, "wire[PRECISION-1:0] topWriteData;\n");
    fprintf(fp, "reg[PRECISION-1:0] topWriteDataLU;\n");
    fprintf(fp, "wire[PRECISION-1:0] topReadData, topReadDataLU;\n");
    fprintf(fp, "wire[TOPSIZEWIDTH-1:0] topWriteAddr, topWriteAddrLU, topReadAddr, topReadAddrLU;\n");
    fprintf(fp, "wire topWriteEn, topWriteEnLU;\n\n");

    if (topoutputdelay != 0) {
        fprintf(fp, "reg[PRECISION-1:0] topReadDataReg[TOPOUTPUTDELAY-1:0];\n");
    }
    if (topinputdelay != 0) {
        fprintf(fp, "reg[PRECISION-1:0] topWriteDataReg[TOPINPUTDELAY-1:0];\n");
        fprintf(fp, "reg[TOPSIZEWIDTH-1:0] topWriteAddrReg[TOPINPUTDELAY-1:0], topReadAddrReg[TOPINPUTDELAY-1:0];\n");
        fprintf(fp, "reg topWriteEnReg[TOPINPUTDELAY-1:0];\n\n");
    }

    fprintf(fp, "wire[RAMWIDTH-1:0] rcWriteData;\n");
    fprintf(fp, "wire leftWriteSel, curWriteSel, topSourceSel;\n");
    fprintf(fp, "wire diagEn;\n");
    fprintf(fp, "wire[PEWIDTH-1:0] topWriteSel;\n\n");

    fprintf(fp, "wire MOSel;\n");
    fprintf(fp, "wire MOEn;\n\n");

    fprintf(fp, "// control block\n");
    fprintf(fp, "LUControl conBlock (clk, start, m, n, loop, mode, done, \n");
    fprintf(fp, "                    curReadAddrLU, curWriteAddrLU, curWriteByteEnLU, curWriteEnLU, curWriteSel,\n");
    fprintf(fp, "                    leftReadAddrLU, leftWriteAddrLU, leftWriteByteEnLU, leftWriteEnLU,  leftWriteSel,\n");
    fprintf(fp, "                    topReadAddrLU, topWriteAddrLU, topWriteEnLU, topWriteSel, topSourceSel, diagEn, MOSel, MOEn);\n\n");
    fprintf(fp, "// fp_div unit\n");

    if (precision == 64) {
        fprintf(fp, "%s rec (clk, 64'h3FF0000000000000, diag, recResult);\n\n", divname);
    }
    else {
        fprintf(fp, "%s rec (clk, 32'h3F800000, diag, recResult);\n\n", divname);
    }

    fprintf(fp, "// on-chip memory blocks that store the matrix to be LU factorized\n");
		fprintf(fp, "// store current blocks data\n");
    fprintf(fp, "ram currentBlock0 (curWriteByteEn0, clk, curWriteData0, curReadAddr0, curWriteAddr0, curWriteEn0, curReadData0);\n");
    fprintf(fp, "ram currentBlock1 (curWriteByteEn1, clk, curWriteData1, curReadAddr1, curWriteAddr1, curWriteEn1, curReadData1);\n");
    fprintf(fp, "// store left blocks data\n");
    fprintf(fp, "ram leftBlock0(leftWriteByteEn0, clk, leftWriteData0, leftReadAddr0, leftWriteAddr0, leftWriteEn0, leftReadData0);\n\n");
    fprintf(fp, "ram leftBlock1(leftWriteByteEn1, clk, leftWriteData1, leftReadAddr1, leftWriteAddr1, leftWriteEn1, leftReadData1);\n\n");
    fprintf(fp, "// store top block data\n");
    fprintf(fp, "top_ram topBlock(clk, topWriteData, topReadAddr, topWriteAddr, topWriteEn, topReadDataLU);\n\n");

    fprintf(fp, "// processing elements that does the main computation of LU factorization\n");
    for (i = 0; i < numPE; i++) {
        fprintf(fp, "mult_add PE%i (clk, multA[%i], multOperand, addA[%i], multResult[%i], addResult[%i]);\n", i, i, i, i, i);
    }
    fprintf(fp, "\n");

    fprintf(fp, "// connect to ports of the left blocks\n");
    fprintf(fp, "assign leftWriteDataLU = (leftWriteSel == 0) ? curReadDataLU : rcWriteData;\n");
    if (meminputdelay == 0)
    {
      fprintf(fp, "assign leftWriteData0 = (leftMemSel == 0) ? leftWriteDataMem : leftWriteDataLU;\n");
      fprintf(fp, "assign leftWriteAddr0 = (leftMemSel == 0) ? leftWriteAddrMem : leftWriteAddrLU;\n");
      fprintf(fp, "assign leftReadAddr0 = leftReadAddrLU;\n");
      fprintf(fp, "assign leftWriteByteEn0 = (leftMemSel == 0) ? leftWriteByteEnMem : leftWriteByteEnLU;\n");
      fprintf(fp, "assign leftWriteEn0 = (leftMemSel == 0) ? leftWriteEnMem : leftWriteEnLU;\n");
      fprintf(fp, "assign leftWriteData1 = (leftMemSel == 0) ? leftWriteDataLU : leftWriteDataMem;\n");
      fprintf(fp, "assign leftWriteAddr1 = (leftMemSel == 0) ? leftWriteAddrLU : leftWriteAddrMem;\n");
      fprintf(fp, "assign leftReadAddr1 = leftReadAddrLU;\n");
      fprintf(fp, "assign leftWriteByteEn1 = (leftMemSel == 0) ? leftWriteByteEnLU : leftWriteByteEnMem;\n");
      fprintf(fp, "assign leftWriteEn1 = (leftMemSel == 0) ? leftWriteEnLU : leftWriteEnMem;\n");
    }
    else
    {
        fprintf(fp, "always @ (posedge clk)\n");
        fprintf(fp, "begin\n");
        fprintf(fp, "	if(leftMemSel == 0)\n");
        fprintf(fp, "	begin\n");
        fprintf(fp, "		leftWriteData0Reg[0] <= leftWriteDataMem;\n");
        fprintf(fp, "		leftWriteAddr0Reg[0] <= leftWriteAddrMem;\n");
        fprintf(fp, "		leftWriteByteEn0Reg[0] <= leftWriteByteEnMem;\n");
        fprintf(fp, "		leftWriteEn0Reg[0] <= leftWriteEnMem;\n");
        fprintf(fp, "		leftWriteData1Reg[0] <= leftWriteDataLU;\n");
        fprintf(fp, "		leftWriteAddr1Reg[0] <= leftWriteAddrLU;\n");
        fprintf(fp, "		leftWriteByteEn1Reg[0] <= leftWriteByteEnLU;\n");
        fprintf(fp, "		leftWriteEn1Reg[0] <= leftWriteEnLU;\n");
        fprintf(fp, "	end\n");
        fprintf(fp, "	else\n");
        fprintf(fp, "	begin\n");
        fprintf(fp, "		leftWriteData0Reg[0] <= leftWriteDataLU;\n");
        fprintf(fp, "		leftWriteAddr0Reg[0] <= leftWriteAddrLU;\n");
        fprintf(fp, "		leftWriteByteEn0Reg[0] <= leftWriteByteEnLU;\n");
        fprintf(fp, "		leftWriteEn0Reg[0] <= leftWriteEnLU;\n");
        fprintf(fp, "		leftWriteData1Reg[0] <= leftWriteDataMem;\n");
        fprintf(fp, "		leftWriteAddr1Reg[0] <= leftWriteAddrMem;\n");
        fprintf(fp, "		leftWriteByteEn1Reg[0] <= leftWriteByteEnMem;\n");
        fprintf(fp, "		leftWriteEn1Reg[0] <= leftWriteEnMem;\n");
        fprintf(fp, "	end\n");
        fprintf(fp, "	leftReadAddr0Reg[0] <= leftReadAddrLU;\n");
        fprintf(fp, "	leftReadAddr1Reg[0] <= leftReadAddrLU;\n");
        for (i = 0; i < meminputdelay-1; i++)
        {
            fprintf(fp, "	leftWriteData0Reg[%i] <= leftWriteData0Reg[%i];\n", i+1, i);
            fprintf(fp, "	leftWriteAddr0Reg[%i] <= leftWriteAddr0Reg[%i];\n", i+1, i);
            fprintf(fp, "	leftReadAddr0Reg[%i] <= leftReadAddr0Reg[%i];\n", i+1, i);
            fprintf(fp, "	leftWriteByteEn0Reg[%i] <= leftWriteByteEn0Reg[%i];\n", i+1, i);
            fprintf(fp, "	leftWriteEn0Reg[%i] <= leftWriteEn0Reg[%i];\n", i+1, i);
            fprintf(fp, "	leftWriteData1Reg[%i] <= leftWriteData1Reg[%i];\n", i+1, i);
            fprintf(fp, "	leftWriteAddr1Reg[%i] <= leftWriteAddr1Reg[%i];\n", i+1, i);
            fprintf(fp, "	leftReadAddr1Reg[%i] <= leftReadAddr1Reg[%i];\n", i+1, i);
            fprintf(fp, "	leftWriteByteEn1Reg[%i] <= leftWriteByteEn1Reg[%i];\n", i+1, i);
            fprintf(fp, "	leftWriteEn1Reg[%i] <= leftWriteEn1Reg[%i];\n", i+1, i);
        }
        fprintf(fp, "end\n");
        fprintf(fp, "assign leftWriteData0 = leftWriteData0Reg[%i];\n", i);
        fprintf(fp, "assign leftWriteAddr0 = leftWriteAddr0Reg[%i];\n", i);
        fprintf(fp, "assign leftReadAddr0 = leftReadAddr0Reg[%i];\n", i);
        fprintf(fp, "assign leftWriteByteEn0 = leftWriteByteEn0Reg[%i];\n", i);
        fprintf(fp, "assign leftWriteEn0 = leftWriteEn0Reg[%i];\n", i);
        fprintf(fp, "assign leftWriteData1 = leftWriteData1Reg[%i];\n", i);
        fprintf(fp, "assign leftWriteAddr1 = leftWriteAddr1Reg[%i];\n", i);
        fprintf(fp, "assign leftReadAddr1 = leftReadAddr1Reg[%i];\n", i);
        fprintf(fp, "assign leftWriteByteEn1 = leftWriteByteEn1Reg[%i];\n", i);
        fprintf(fp, "assign leftWriteEn1 = leftWriteEn1Reg[%i];\n", i);
    }
    fprintf(fp, "\n");

    if (memoutputdelay == 0) {
        fprintf(fp, "assign leftReadDataLU = (leftMemSel == 0) ? leftReadData1 : leftReadData0;\n");
    }
    else {
        fprintf(fp, "always @ (posedge clk)\n");
        fprintf(fp, "begin\n");
        fprintf(fp, "		leftReadData0Reg[0] <= leftReadData0;\n");
        fprintf(fp, "		leftReadData1Reg[0] <= leftReadData1;\n");
        for (i = 0; i < memoutputdelay-1; i++)
        {
            fprintf(fp, "	leftReadData0Reg[%i] <= leftReadData0Reg[%i];\n", i+1, i);
            fprintf(fp, "	leftReadData1Reg[%i] <= leftReadData1Reg[%i];\n", i+1, i);
        }
        fprintf(fp, "end\n");
        fprintf(fp, "assign leftReadDataLU = (leftMemSel == 0) ? leftReadData1Reg[%i] : leftReadData0Reg[%i];\n", i, i);
    }

    fprintf(fp, "// data feed to fp div unit\n");
    fprintf(fp, "always @ (posedge clk)\n");
    fprintf(fp, "begin\n");
    fprintf(fp, "	if (diagEn == 1)\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "			diag <= topReadData;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "end\n");

    fprintf(fp, "// one of the inputs to the PE\n");
    fprintf(fp, "always @ (posedge clk)\n");
    fprintf(fp, "begin\n");
    fprintf(fp, "	if (start == 1)\n");
    fprintf(fp, "		multOperand <= 0;\n");
    fprintf(fp, "	else if (MOEn == 1)\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		if (MOSel == 0)\n");
    fprintf(fp, "			multOperand <= recResult;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			multOperand <= topReadData;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "end\n\n");

    fprintf(fp, "// connections to top block memory ports\n");
    fprintf(fp, "always @ (*)\n");
    fprintf(fp, "begin\n");
    fprintf(fp, "	if (topSourceSel == 0)\n");
    fprintf(fp, "		case (topWriteSel)\n");
    for (i = 0; i < numPE; i++)
    {
        lowerIdx = (numPE-i-1)*precision;
        upperIdx = (numPE-i)*precision-1;
        fprintf(fp, "		%i:\n", i);
        fprintf(fp, "			topWriteDataLU = curReadDataLU[%i:%i];\n", upperIdx, lowerIdx);
    }
    fprintf(fp, "		default:\n");
    fprintf(fp, "			topWriteDataLU = curReadDataLU[PRECISION-1:0];\n");
    fprintf(fp, "		endcase\n");
    fprintf(fp, "	else\n");
    fprintf(fp, "		case (topWriteSel)\n");
    for (i = 0; i < numPE; i++)
    {
        fprintf(fp, "		%i:\n", i);
        fprintf(fp, "			topWriteDataLU = addResult[%i];\n", numPE-i-1);
    }
    fprintf(fp, "		default:\n");
    fprintf(fp, "			topWriteDataLU = addResult[0];\n");
    fprintf(fp, "		endcase\n");
    fprintf(fp, "end\n\n");

    if (topinputdelay == 0) {
        fprintf(fp, "assign topWriteData = topWriteDataLU;\n");
        fprintf(fp, "assign topReadAddr = topReadAddrLU;\n");
        fprintf(fp, "assign topWriteAddr = topWriteAddrLU;\n");
        fprintf(fp, "assign topWriteEn = topWriteEnLU;\n");
    }
    else {
        fprintf(fp, "always @ (posedge clk)\n");
        fprintf(fp, "begin\n");
        fprintf(fp, "	topWriteDataReg[0] <= topWriteDataLU;\n");
        fprintf(fp, "	topReadAddrReg[0] <= topReadAddrLU;\n");
        fprintf(fp, "	topWriteAddrReg[0] <= topWriteAddrLU;\n");
        fprintf(fp, "	topWriteEnReg[0] <= topWriteEnLU;\n");
        for (i = 0; i < topinputdelay-1; i++) {
            fprintf(fp, "	topWriteDataReg[%i] <= topWriteDataReg[%i];\n", i+1, i);
            fprintf(fp, "	topReadAddrReg[%i] <= topReadAddrReg[%i];\n", i+1, i);
            fprintf(fp, "	topWriteAddrReg[%i] <= topWriteAddrReg[%i];\n", i+1, i);
            fprintf(fp, "	topWriteEnReg[%i] <= topWriteEnReg[%i];\n", i+1, i);
        }
        fprintf(fp, "end\n");
        fprintf(fp, "assign topWriteData = topWriteDataReg[%i];\n", i);
        fprintf(fp, "assign topReadAddr = topReadAddrReg[%i];\n", i);
        fprintf(fp, "assign topWriteAddr = topWriteAddrReg[%i];\n", i);
        fprintf(fp, "assign topWriteEn = topWriteEnReg[%i];\n", i);
    }
    if (topoutputdelay == 0) {
        fprintf(fp, "assign topReadData = topReadDataLU;\n");
    }
    else {
        fprintf(fp, "always @ (posedge clk)\n");
        fprintf(fp, "begin\n");
        fprintf(fp, "	topReadDataReg[0] <= topReadDataLU;\n");
        for (i = 0; i < topoutputdelay-1; i++) {
            fprintf(fp, "	topReadDataReg[%i] <= topReadDataReg[%i];\n", i+1, i);
        }
        fprintf(fp, "end\n");
        fprintf(fp, "assign topReadData = topReadDataReg[%i];\n", i);
    }
    fprintf(fp, "\n");

    fprintf(fp, "// connections to processing element\n");
    for (i = 0; i < numPE; i++)
    {
        lowerIdx = i*precision;
        upperIdx = (i+1)*precision-1;
        fprintf(fp, "assign multA[%i] = leftReadDataLU[%i:%i];\n", i, upperIdx, lowerIdx);
    }
    fprintf(fp, "\n");

    for (i = 0; i < numPE; i++)
    {
        lowerIdx = i*precision;
        upperIdx = (i+1)*precision-1;
        fprintf(fp, "assign addA[%i] = curReadDataLU[%i:%i];\n", i, upperIdx, lowerIdx);
    }
    fprintf(fp, "\n");

    fprintf(fp, "// connections to ports of the current blocks\n");
    for (i = 0; i < numPE; i++)
    {
        lowerIdx = i*precision;
        upperIdx = (i+1)*precision-1;
        fprintf(fp, "assign rcWriteData[%i:%i] = (curWriteSel == 0) ? multResult[%i] : addResult[%i];\n", upperIdx, lowerIdx, i, i);
    }
    fprintf(fp, "assign curWriteDataLU = rcWriteData;\n");
    fprintf(fp, "\n");

    if (meminputdelay == 0)
    {
      fprintf(fp, "assign curWriteData0 = (curMemSel == 0) ? curWriteDataMem : curWriteDataLU;\n");
      fprintf(fp, "assign curWriteAddr0 = (curMemSel == 0) ? curWriteAddrMem : curWriteAddrLU;\n");
      fprintf(fp, "assign curReadAddr0 = (curMemSel == 0) ? curReadAddrMem : curReadAddrLU;\n");
      fprintf(fp, "assign curWriteByteEn0 = (curMemSel == 0) ? curWriteByteEnMem : curWriteByteEnLU;\n");
      fprintf(fp, "assign curWriteEn0 = (curMemSel == 0) ? curWriteEnMem : curWriteEnLU;\n");
      fprintf(fp, "assign curWriteData1 = (curMemSel == 0) ? curWriteDataLU : curWriteDataMem;\n");
      fprintf(fp, "assign curWriteAddr1 = (curMemSel == 0) ? curWriteAddrLU : curWriteAddrMem;\n");
      fprintf(fp, "assign curReadAddr1 = (curMemSel == 0) ? curReadAddrLU : curReadAddrMem;\n");
      fprintf(fp, "assign curWriteByteEn1 = (curMemSel == 0) ? curWriteByteEnLU : curWriteByteEnMem;\n");
      fprintf(fp, "assign curWriteEn1 = (curMemSel == 0) ? curWriteEnLU : curWriteEnMem;\n");
    }
    else
    {
        fprintf(fp, "always @ (posedge clk)\n");
        fprintf(fp, "begin\n");
        fprintf(fp, "	if(curMemSel == 0)\n");
        fprintf(fp, "	begin\n");
        fprintf(fp, "		curWriteData0Reg[0] <= curWriteDataMem;\n");
        fprintf(fp, "		curWriteAddr0Reg[0] <= curWriteAddrMem;\n");
        fprintf(fp, "		curReadAddr0Reg[0] <= curReadAddrMem;\n");
        fprintf(fp, "		curWriteByteEn0Reg[0] <= curWriteByteEnMem;\n");
        fprintf(fp, "		curWriteEn0Reg[0] <= curWriteEnMem;\n");
        fprintf(fp, "		curWriteData1Reg[0] <= curWriteDataLU;\n");
        fprintf(fp, "		curWriteAddr1Reg[0] <= curWriteAddrLU;\n");
        fprintf(fp, "		curReadAddr1Reg[0] <= curReadAddrLU;\n");
        fprintf(fp, "		curWriteByteEn1Reg[0] <= curWriteByteEnLU;\n");
        fprintf(fp, "		curWriteEn1Reg[0] <= curWriteEnLU;\n");
        fprintf(fp, "	end\n");
        fprintf(fp, "	else\n");
        fprintf(fp, "	begin\n");
        fprintf(fp, "		curWriteData0Reg[0] <= curWriteDataLU;\n");
        fprintf(fp, "		curWriteAddr0Reg[0] <= curWriteAddrLU;\n");
        fprintf(fp, "		curReadAddr0Reg[0] <= curReadAddrLU;\n");
        fprintf(fp, "		curWriteByteEn0Reg[0] <= curWriteByteEnLU;\n");
        fprintf(fp, "		curWriteEn0Reg[0] <= curWriteEnLU;\n");
        fprintf(fp, "		curWriteData1Reg[0] <= curWriteDataMem;\n");
        fprintf(fp, "		curWriteAddr1Reg[0] <= curWriteAddrMem;\n");
        fprintf(fp, "		curReadAddr1Reg[0] <= curReadAddrMem;\n");
        fprintf(fp, "		curWriteByteEn1Reg[0] <= curWriteByteEnMem;\n");
        fprintf(fp, "		curWriteEn1Reg[0] <= curWriteEnMem;\n");
        fprintf(fp, "	end\n");
        for (i = 0; i < meminputdelay-1; i++)
        {
            fprintf(fp, "	curWriteData0Reg[%i] <= curWriteData0Reg[%i];\n", i+1, i);
            fprintf(fp, "	curWriteAddr0Reg[%i] <= curWriteAddr0Reg[%i];\n", i+1, i);
            fprintf(fp, "	curReadAddr0Reg[%i] <= curReadAddr0Reg[%i];\n", i+1, i);
            fprintf(fp, "	curWriteByteEn0Reg[%i] <= curWriteByteEn0Reg[%i];\n", i+1, i);
            fprintf(fp, "	curWriteEn0Reg[%i] <= curWriteEn0Reg[%i];\n", i+1, i);
            fprintf(fp, "	curWriteData1Reg[%i] <= curWriteData1Reg[%i];\n", i+1, i);
            fprintf(fp, "	curWriteAddr1Reg[%i] <= curWriteAddr1Reg[%i];\n", i+1, i);
            fprintf(fp, "	curReadAddr1Reg[%i] <= curReadAddr1Reg[%i];\n", i+1, i);
            fprintf(fp, "	curWriteByteEn1Reg[%i] <= curWriteByteEn1Reg[%i];\n", i+1, i);
            fprintf(fp, "	curWriteEn1Reg[%i] <= curWriteEn1Reg[%i];\n", i+1, i);
        }
        fprintf(fp, "end\n");
        fprintf(fp, "assign curWriteData0 = curWriteData0Reg[%i];\n", i);
        fprintf(fp, "assign curWriteAddr0 = curWriteAddr0Reg[%i];\n", i);
        fprintf(fp, "assign curReadAddr0 = curReadAddr0Reg[%i];\n", i);
        fprintf(fp, "assign curWriteByteEn0 = curWriteByteEn0Reg[%i];\n", i);
        fprintf(fp, "assign curWriteEn0 = curWriteEn0Reg[%i];\n", i);
        fprintf(fp, "assign curWriteData1 = curWriteData1Reg[%i];\n", i);
        fprintf(fp, "assign curWriteAddr1 = curWriteAddr1Reg[%i];\n", i);
        fprintf(fp, "assign curReadAddr1 = curReadAddr1Reg[%i];\n", i);
        fprintf(fp, "assign curWriteByteEn1 = curWriteByteEn1Reg[%i];\n", i);
        fprintf(fp, "assign curWriteEn1 = curWriteEn1Reg[%i];\n", i);
    }
    fprintf(fp, "\n");

    if (memoutputdelay == 0) {
        fprintf(fp, "assign curReadDataMem = (curMemSel == 0) ? curReadData0 : curReadData1;\n");
        fprintf(fp, "assign curReadDataLU = (curMemSel == 0) ? curReadData1 : curReadData0;\n");
    }
    else {
        fprintf(fp, "always @ (posedge clk)\n");
        fprintf(fp, "begin\n");
        fprintf(fp, "		curReadData0Reg[0] <= curReadData0;\n");
        fprintf(fp, "		curReadData1Reg[0] <= curReadData1;\n");
        for (i = 0; i < memoutputdelay-1; i++)
        {
            fprintf(fp, "	curReadData0Reg[%i] <= curReadData0Reg[%i];\n", i+1, i);
            fprintf(fp, "	curReadData1Reg[%i] <= curReadData1Reg[%i];\n", i+1, i);
        }
        fprintf(fp, "end\n");
        fprintf(fp, "assign curReadDataMem = (curMemSel == 0) ? curReadData0Reg[%i] : curReadData1Reg[%i];\n", i, i);
        fprintf(fp, "assign curReadDataLU = (curMemSel == 0) ? curReadData1Reg[%i] : curReadData0Reg[%i];\n", i, i);
    }

    fprintf(fp, "endmodule\n");
    fclose(fp);
}
