    // generate DataTransferUnit module
    // by Wei Zhang
    //

#include <stdio.h>
#include <math.h>
#define intlog2(x) (int)ceil(log(x)/log(2))

void genDTU(int burstlen, int datawidth, int memconwidth, int ddrsizewidth, int fifosize,
    int ramwidth, int ramsizewidth, int ratio, int ramlat, int transfersize)
{
    FILE *fp;
    int i, j, k;
    int burstwidth, fifosizewidth, countwidth;

    burstwidth = intlog2(burstlen+1);
    countwidth = intlog2(ratio+burstlen);
    fifosizewidth = intlog2(fifosize);

    fp = fopen("DataTransferUnit.v", "w");

    fprintf(fp, "\n");
    fprintf(fp, "module DataTransferUnit (clk, phy_clk, dtu_write_req, dtu_read_req, dtu_mem_addr, dtu_ram_addr, dtu_size, dtu_ack, dtu_done,\n");
    fprintf(fp, "		ram_read_addr, ram_read_data, ram_write_byte_en, ram_write_data, ram_write_addr, ram_write_en,\n");
    fprintf(fp, "		mem_rdata, mem_rdata_valid, mem_ready, mem_wdata_req, reset_n,\n");
    fprintf(fp, "		burst_begin, mem_local_addr, mem_be, mem_read_req, mem_size, mem_wdata, mem_write_req);\n");
    fprintf(fp, "\n");
    fprintf(fp, "parameter BURSTLEN = %i, BURSTWIDTH = %i;\n", burstlen, burstwidth);
    fprintf(fp, "parameter DATAWIDTH = %i, DATANUMBYTES = %i;\n", datawidth, datawidth/8);
    fprintf(fp, "parameter MEMCONWIDTH = %i, MEMCONNUMBYTES = %i, DDRSIZEWIDTH = %i;\n", memconwidth, memconwidth/8, ddrsizewidth);
    fprintf(fp, "parameter FIFOSIZE = %i, FIFOSIZEWIDTH = %i;\n", fifosize, fifosizewidth);
    fprintf(fp, "parameter RAMWIDTH = %i, RAMNUMBYTES = %i, RAMSIZEWIDTH = %i;\n", ramwidth, ramwidth/8, ramsizewidth);
    fprintf(fp, "parameter RATIO = %i, RAMLAT = %i;\n", ratio, ramlat);
    fprintf(fp, "\n");
    fprintf(fp, "output burst_begin;\n");
    fprintf(fp, "output [DDRSIZEWIDTH-1:0] mem_local_addr;\n");
    fprintf(fp, "output [MEMCONNUMBYTES-1: 0] mem_be;\n");
    fprintf(fp, "output mem_read_req;\n");
    fprintf(fp, "output [BURSTWIDTH-1:0] mem_size;\n");
    fprintf(fp, "output [MEMCONWIDTH-1:0] mem_wdata;\n");
    fprintf(fp, "output mem_write_req;\n");
    fprintf(fp, "input clk, phy_clk;\n");
    fprintf(fp, "input [MEMCONWIDTH-1:0] mem_rdata;\n");
    fprintf(fp, "input mem_rdata_valid;\n");
    fprintf(fp, "input mem_ready;\n");
    fprintf(fp, "input mem_wdata_req;\n");
    fprintf(fp, "input reset_n;\n");
    fprintf(fp, "\n");
    fprintf(fp, "input dtu_write_req;\n");
    fprintf(fp, "input dtu_read_req;\n");
    fprintf(fp, "input [DDRSIZEWIDTH-1:0] dtu_mem_addr;\n");
    fprintf(fp, "input [RAMSIZEWIDTH-1:0] dtu_ram_addr;\n");
    fprintf(fp, "input [%i:0] dtu_size;\n", transfersize-1);
    fprintf(fp, "output dtu_ack;\n");
    fprintf(fp, "output dtu_done;\n");
    fprintf(fp, "\n");
    fprintf(fp, "output[RAMWIDTH-1:0] ram_write_data;\n");
    fprintf(fp, "input[RAMWIDTH-1:0] ram_read_data;\n");
    fprintf(fp, "output[RAMSIZEWIDTH-1:0] ram_write_addr, ram_read_addr;\n");
    fprintf(fp, "output[RAMNUMBYTES-1:0] ram_write_byte_en;\n");
    fprintf(fp, "output ram_write_en;\n");
    fprintf(fp, "\n");
    fprintf(fp, "parameter IDLE = 0, WRITE = 1, READ = 2;\n");
    fprintf(fp, "reg [DDRSIZEWIDTH-1:0] mem_addr[RAMLAT:0];\n");
    fprintf(fp, "reg [2:0] state;\n");
    fprintf(fp, "wire [DATAWIDTH-1:0] rdata, ram_write_dataw, ram_read_dataw;\n");
    fprintf(fp, "\n");
    fprintf(fp, "wire [RAMSIZEWIDTH-1:0] rfifo_addr;\n");
    fprintf(fp, "reg fifo_write_reg[RAMLAT-1:0];\n");
    fprintf(fp, "reg write_req_reg[RAMLAT-1:0];\n");
    fprintf(fp, "reg read_req_reg[RAMLAT-1:0];\n");
    fprintf(fp, "reg fifo_read_reg[0:0];\n");
    fprintf(fp, "reg rdata_valid;\n");
    fprintf(fp, "reg test_complete_reg[1:0];\n");
    fprintf(fp, "reg [BURSTWIDTH-1:0] size_count[RAMLAT-1:0];\n");
    fprintf(fp, "reg [RAMSIZEWIDTH-1:0] size;\n");
    fprintf(fp, "reg [RAMSIZEWIDTH-1:0] ram_addr[RAMLAT-1:0];\n");
    fprintf(fp, "reg [%i:0] data_count;\n", countwidth-1);
    fprintf(fp, "reg ram_write_en_reg;\n");
    fprintf(fp, "\n");
    fprintf(fp, "wire read_req;\n");
    fprintf(fp, "wire write_req;\n");
    fprintf(fp, "wire [FIFOSIZEWIDTH-1:0] wfifo_count;\n");
    fprintf(fp, "wire rfull, wempty, rempty, rdcmd_empty, wrcmd_full, wrcmd_empty, rdata_empty;\n");
    fprintf(fp, "wire [DATAWIDTH-1:0] mem_data;\n");
    fprintf(fp, "wire not_stall;\n");
    fprintf(fp, "wire fifo_write, fifo_read;\n");
    fprintf(fp, "wire rdata_req;\n");
    fprintf(fp, "wire [BURSTWIDTH+DDRSIZEWIDTH+1:0] wrmem_cmd, rdmem_cmd;\n");
    fprintf(fp, "wire mem_cmd_ready, mem_cmd_issue;\n");
    fprintf(fp, "\n");

    fprintf(fp, "// FIFOs to interact with off-chip memory\n");
    fprintf(fp, "memcmd_fifo cmd_store(\n");
    fprintf(fp, "	.aclr(~reset_n),\n");
    fprintf(fp, "	.rdclk(phy_clk),\n");
    fprintf(fp, "	.wrclk(clk),\n");
    fprintf(fp, "	.data(wrmem_cmd),\n");
    fprintf(fp, "	.rdreq(mem_cmd_ready),\n");
    fprintf(fp, "	.rdempty(rdcmd_empty),\n");
    fprintf(fp, "	.wrreq(mem_cmd_issue),\n");
    fprintf(fp, "	.wrfull(wrcmd_full),\n");
    fprintf(fp, "	.wrempty(wrcmd_empty),\n");
    fprintf(fp, "	.q(rdmem_cmd));\n");
    fprintf(fp, "\n");
    fprintf(fp, "wfifo wdata_store(\n");
    fprintf(fp, "	.rdclk(phy_clk),\n");
    fprintf(fp, "	.wrclk(clk),\n");
    fprintf(fp, "	.data(mem_data),\n");
    fprintf(fp, "	.rdreq(mem_wdata_req),\n");
    fprintf(fp, "	.wrreq(fifo_write),\n");
    fprintf(fp, "	.wrempty(wempty),\n");
    fprintf(fp, "	.q(mem_wdata),\n");
    fprintf(fp, "	.wrusedw(wfifo_count));\n");
    fprintf(fp, "\n");
    fprintf(fp, "addr_fifo raddress_store (\n");
    fprintf(fp, "	.clock(clk),\n");
    fprintf(fp, "	.data(ram_addr[RAMLAT-2]),\n");
    fprintf(fp, "	.rdreq(rdata_req),\n");
    fprintf(fp, "	.wrreq(fifo_read),\n");
    fprintf(fp, "	.empty(rempty),\n");
    fprintf(fp, "	.full(rfull),\n");
    fprintf(fp, "	.q(rfifo_addr));\n");
    fprintf(fp, "\n");
    fprintf(fp, "rfifo rdata_store(\n");
    fprintf(fp, "	.data(mem_rdata),\n");
    fprintf(fp, "	.rdclk(clk),\n");
    fprintf(fp, "	.rdreq(rdata_req),\n");
    fprintf(fp, "	.wrclk(phy_clk),\n");
    fprintf(fp, "	.wrreq(mem_rdata_valid),\n");
    fprintf(fp, "	.q(rdata),\n");
    fprintf(fp, "	.rdempty(rdata_empty));\n");
    fprintf(fp, "\n");

    fprintf(fp, "assign mem_cmd_ready = (mem_ready == 1) && (rdcmd_empty == 0);\n");
    fprintf(fp, "assign mem_cmd_issue = (wrcmd_full == 0) && (write_req == 1 || read_req == 1 || wrcmd_empty == 1);\n");
    fprintf(fp, "assign wrmem_cmd[BURSTWIDTH+DDRSIZEWIDTH+1:DDRSIZEWIDTH+2] = size_count[0];\n");
    fprintf(fp, "assign wrmem_cmd[DDRSIZEWIDTH+1:2] = mem_addr[0];\n");
    fprintf(fp, "assign wrmem_cmd[1] = read_req;\n");
    fprintf(fp, "assign wrmem_cmd[0] = write_req;\n");
    fprintf(fp, "assign mem_write_req = rdmem_cmd[0] && rdcmd_empty == 0;\n");
    fprintf(fp, "assign mem_read_req = rdmem_cmd[1] && rdcmd_empty == 0;\n");
    fprintf(fp, "assign mem_local_addr = rdmem_cmd[DDRSIZEWIDTH+1:2];\n");
    fprintf(fp, "assign burst_begin = 0;\n");
    fprintf(fp, "assign mem_size = rdmem_cmd[BURSTWIDTH+DDRSIZEWIDTH+1:DDRSIZEWIDTH+2];\n");
    fprintf(fp, "assign mem_be = ~0;\n");
    fprintf(fp, "assign fifo_write = fifo_write_reg[0];\n");
    fprintf(fp, "assign write_req = (not_stall) ? write_req_reg[0] : 0;\n");
    fprintf(fp, "assign read_req = (not_stall) ? read_req_reg[0] : 0;\n");
    fprintf(fp, "assign fifo_read = (not_stall) ? fifo_read_reg[0] : 0;\n");
    fprintf(fp, "assign not_stall = (wfifo_count < FIFOSIZE-%i) && (rfull == 0) && (wrcmd_full == 0);\n", ramlat);
    fprintf(fp, "assign dtu_ack = (state == IDLE);\n");
    fprintf(fp, "assign dtu_done = (state == IDLE) && wempty && rempty;\n");
    fprintf(fp, "\n");
    // match how memory is stored (like switching endianness)
    j = memconwidth;
    k = datawidth;
    for (i = 0; i < ratio; i++)
    {
        fprintf(fp, "assign ram_write_dataw[%i:%i] = rdata[%i:%i];\n", j-1, j-memconwidth, k-1, k-memconwidth);
        fprintf(fp, "assign mem_data[%i:%i] = ram_read_dataw[%i:%i];\n", j-1, j-memconwidth, k-1, k-memconwidth);
        j += memconwidth;
        k -= memconwidth;
    }
    fprintf(fp, "assign ram_write_data = ram_write_dataw[%i:%i];\n", datawidth-1, datawidth-ramwidth);
    fprintf(fp, "assign ram_read_dataw[%i:%i] = ram_read_data;\n", datawidth-1, datawidth-ramwidth);
    if (datawidth != ramwidth) {
        fprintf(fp, "assign ram_read_dataw[%i:0] = 0;\n", datawidth-ramwidth-1);
    }
    fprintf(fp, "assign ram_write_addr = rfifo_addr;\n");
    fprintf(fp, "assign ram_read_addr = ram_addr[RAMLAT-1];\n");
    fprintf(fp, "assign ram_write_byte_en = ~0;\n");
    fprintf(fp, "assign ram_write_en = ram_write_en_reg;\n");
    fprintf(fp, "assign rdata_req = !rdata_empty;\n");
    fprintf(fp, "\n");

    fprintf(fp, "// FSM to produce off-chip memory commands\n");
    fprintf(fp, "always @ (posedge clk)\n");
    fprintf(fp, "begin\n");
    fprintf(fp, "	if (reset_n == 0)\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		state <= IDLE;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	else\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		case (state)\n");
    fprintf(fp, "		IDLE:\n");
    fprintf(fp, "		begin\n");
    fprintf(fp, "			if (dtu_write_req)\n");
    fprintf(fp, "				state <= WRITE;\n");
    fprintf(fp, "			else if (dtu_read_req)\n");
    fprintf(fp, "				state <= READ;\n");
    fprintf(fp, "			else\n");
    fprintf(fp, "				state <= IDLE;\n");
    fprintf(fp, "		end\n");
    fprintf(fp, "		WRITE:\n");
    fprintf(fp, "		begin\n");
    fprintf(fp, "			if (not_stall && size == 0 && data_count < BURSTLEN)\n");
    fprintf(fp, "				state <= IDLE;\n");
    fprintf(fp, "			else\n");
    fprintf(fp, "				state <= WRITE;\n");
    fprintf(fp, "		end\n");
    fprintf(fp, "		READ:\n");
    fprintf(fp, "		begin\n");
    fprintf(fp, "			if (not_stall && size == 0 && data_count < BURSTLEN)\n");
    fprintf(fp, "				state <= IDLE;\n");
    fprintf(fp, "			else\n");
    fprintf(fp, "				state <= READ;\n");
    fprintf(fp, "		end\n");
    fprintf(fp, "		default:\n");
    fprintf(fp, "		begin\n");
    fprintf(fp, "			state <= IDLE;\n");
    fprintf(fp, "		end\n");
    fprintf(fp, "		endcase\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "end\n");
    fprintf(fp, "\n");

    fprintf(fp, "always @ (posedge clk)\n");
    fprintf(fp, "begin\n");
    fprintf(fp, "\n");
    fprintf(fp, "	if (reset_n == 0)\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		size <= 0;\n");
    fprintf(fp, "		data_count <= 0;\n");
    fprintf(fp, "		size_count[RAMLAT-1] <= 1;\n");
    fprintf(fp, "		mem_addr[RAMLAT] <= 0;\n");
    fprintf(fp, "		ram_addr[RAMLAT-1] <= 0;\n");
    fprintf(fp, "		fifo_write_reg[RAMLAT-1] <= 0;\n");
    fprintf(fp, "		write_req_reg[RAMLAT-1] <= 0;\n");
    fprintf(fp, "		fifo_read_reg[0] <= 0;\n");
    fprintf(fp, "		read_req_reg[RAMLAT-1] <= 0;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	else if (state == IDLE)\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		size <= dtu_size;\n");
    fprintf(fp, "		size_count[RAMLAT-1] <= BURSTLEN;\n");
    fprintf(fp, "		mem_addr[RAMLAT] <= dtu_mem_addr;\n");
    fprintf(fp, "		ram_addr[RAMLAT-1] <= dtu_ram_addr;\n");
    fprintf(fp, "		fifo_write_reg[RAMLAT-1] <= 0;\n");
    fprintf(fp, "		write_req_reg[RAMLAT-1] <= 0;\n");
    fprintf(fp, "		fifo_read_reg[0] <= 0;\n");
    fprintf(fp, "		read_req_reg[RAMLAT-1] <= 0;\n");
    fprintf(fp, "		data_count <= 0;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	else if (data_count >= BURSTLEN && not_stall)\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		data_count <= data_count - BURSTLEN;\n");
    fprintf(fp, "		mem_addr[RAMLAT] <= mem_addr[RAMLAT] + BURSTLEN;\n");
    fprintf(fp, "		fifo_write_reg[RAMLAT-1] <= 0;\n");
    fprintf(fp, "		write_req_reg[RAMLAT-1] <= state == WRITE;\n");
    fprintf(fp, "		fifo_read_reg[0] <= 0;\n");
    fprintf(fp, "		read_req_reg[RAMLAT-1] <= state == READ;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	else if (size == 0 && data_count == 0 && not_stall)\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		fifo_write_reg[RAMLAT-1] <= 0;\n");
    fprintf(fp, "		write_req_reg[RAMLAT-1] <= 0;\n");
    fprintf(fp, "		fifo_read_reg[0] <= 0;\n");
    fprintf(fp, "		read_req_reg[RAMLAT-1] <= 0;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	else if (size == 0 && not_stall)\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		size_count[RAMLAT-1] <= data_count[BURSTWIDTH-1:0];\n");
    fprintf(fp, "		fifo_write_reg[RAMLAT-1] <= 0;\n");
    fprintf(fp, "		write_req_reg[RAMLAT-1] <= state == WRITE;\n");
    fprintf(fp, "		fifo_read_reg[0] <= 0;\n");
    fprintf(fp, "		read_req_reg[RAMLAT-1] <= state == READ;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	else if (not_stall)\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		size <= size - 1;\n");
    fprintf(fp, "		data_count <= data_count + RATIO - BURSTLEN;\n");
    fprintf(fp, "		mem_addr[RAMLAT] <= mem_addr[RAMLAT] + BURSTLEN;\n");
    fprintf(fp, "		ram_addr[RAMLAT-1] <= ram_addr[RAMLAT-1]+1;\n");
    fprintf(fp, "		fifo_write_reg[RAMLAT-1] <= state == WRITE;\n");
    fprintf(fp, "		write_req_reg[RAMLAT-1] <= state == WRITE;\n");
    fprintf(fp, "		fifo_read_reg[0] <= state == READ;\n");
    fprintf(fp, "		read_req_reg[RAMLAT-1] <= state == READ;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	else\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		fifo_write_reg[RAMLAT-1] <= 0;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "end\n");
    fprintf(fp, "\n");
    fprintf(fp, "\n");

    fprintf(fp, "always @ (posedge clk)\n");
    fprintf(fp, "begin\n");
    fprintf(fp, "	if (reset_n == 0)\n");
    fprintf(fp, "	begin\n");
    for (i = 0; i < ramlat-1; i++)
    {
        fprintf(fp, "		fifo_write_reg[%i] <= 0;\n", i);
    }
    fprintf(fp, "	end\n");
    fprintf(fp, "	else\n");
    fprintf(fp, "	begin\n");
    for (i = 0; i < ramlat-1; i++)
    {
        fprintf(fp, "		fifo_write_reg[%i] <= fifo_write_reg[%i];\n", i, i+1);
    }
    fprintf(fp, "	end\n");
    fprintf(fp, "\n");
    fprintf(fp, "	if (reset_n == 0)\n");
    fprintf(fp, "	begin\n");
    for (i = 0; i < ramlat-1; i++)
    {
        fprintf(fp, "		mem_addr[%i] <= 0;\n", i);
        fprintf(fp, "		ram_addr[%i] <= 0;\n", i);
        fprintf(fp, "		size_count[%i] <= 1;\n", i);
        fprintf(fp, "		write_req_reg[%i] <= 0;\n", i);
        fprintf(fp, "		read_req_reg[%i] <= 0;\n", i);
    }
    fprintf(fp, "		mem_addr[%i] <= 0;\n", ramlat-1);
    fprintf(fp, "	end\n");
    fprintf(fp, "	else if (not_stall)\n");
    fprintf(fp, "	begin\n");
    for (i = 0; i < ramlat-1; i++)
    {
        fprintf(fp, "		size_count[%i] <= size_count[%i];\n", i, i+1);
        fprintf(fp, "		mem_addr[%i] <= mem_addr[%i];\n", i, i+1);
        fprintf(fp, "		ram_addr[%i] <= ram_addr[%i];\n", i, i+1);
        fprintf(fp, "		write_req_reg[%i] <= write_req_reg[%i];\n", i, i+1);
        fprintf(fp, "		read_req_reg[%i] <= read_req_reg[%i];\n", i, i+1);
    }
    fprintf(fp, "		mem_addr[%i] <= mem_addr[%i];\n", ramlat-1, ramlat);
    fprintf(fp, "	end\n");
    fprintf(fp, "	\n");
    fprintf(fp, "	ram_write_en_reg  <= rdata_req;\n");
    fprintf(fp, "end\n");
    fprintf(fp, "\n");
    fprintf(fp, "endmodule\n");
    fclose(fp);
}
