#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <sys/stat.h>

#define intlog2(x) (int)ceil(log(x)/log(2))

static char corelibname[100] = "CoreLib";

int copyfile(char *srcname, char *dstname)
{
    FILE *sfp, *dfp;
    char ch;

    // open source file
    sfp = fopen(srcname, "rb");
    if(sfp == NULL) 
    {
        printf("Erorr: copying file - cannot open input file \"%s\"\n", srcname);
        return 1;
    }

    // open destination file
    dfp = fopen(dstname, "wb");
    if(dfp == NULL) 
    {
        printf("Erorr: copying file - cannot open output file \"%s\"\n", dstname);
        fclose(sfp);
        return 1;
    }

    // copy data
    while(!feof(sfp)) 
    {
        ch = getc(sfp);
        if(ferror(sfp)) {
            printf("Error: copying file - cannot read input file \"%s\"\n", srcname);
            fclose(sfp);
            fclose(dfp);
            return 1;
        } 
        else 
        {
            if(!feof(sfp)) 
                putc(ch, dfp);
            if(ferror(dfp)) {
                printf("Error: copying file - cannot write to output file \"%s\"\n", dstname);
                fclose(sfp);
                fclose(dfp);
                return 1;
            }
        }
    }

    fclose(sfp);
    fclose(dfp);
    return 0;
}

int copydir(char *srcdir, char *dstdir)
{
    DIR *dp, *tempdp;
    struct dirent *dentry;
    char srcname[500], dstname[500];
    int len;

    // open source directory
    dp = opendir(srcdir);
    if (dp == NULL)
    {
                printf("Error: cannot find core library directory %s\n", srcdir);
                return 1;
    }
 
    printf("copying files in %s\n", srcdir);

    while ((dentry = readdir(dp)) != NULL)
    {
        // check each item in directory
        len = strlen(srcdir);
        strcpy(srcname, srcdir);
        srcname[len] = '/';
        strcpy(srcname+len+1, dentry->d_name);
        tempdp = opendir(srcname);
        if (tempdp != NULL)
        {
            // is a director, do not copy
            closedir(tempdp);
        }
        else
        {
            // not a directory (then a file), copy file
            len = strlen(dstdir);
            strcpy(dstname, dstdir);
            dstname[len] = '/';
            strcpy(dstname+len+1, dentry->d_name);
            //printf("srcname %s, dstname %s\n", srcname, dstname);
            len = copyfile(srcname, dstname);
            if(len == 1)
            {
                printf("Error: copying directory %s\n", srcdir);
                return 1;
            }
        }
    }
    closedir(dp);
    return 0;
}

int checkVal(char * parameter, char * valptr, int *value, int condition)
{
    // condition = 0 means can only be any non-negative integer (0, 1, 2, ...)
    // condition = 1 means can only be any positive integer (1, 2, 3, ...)
    // condition = 2 means can only be power of 2 integer (1, 2, 4, ...)

    char *ptr;
    int temp;

    // check if value is non-negative integer
    if (*valptr < '0' || *valptr > '9')
    {
        ptr = valptr;
        if (*ptr == '\n' || *ptr == '\r' || *ptr == '#')
            printf("Error: missing %s parameter value in assignment\n", parameter);
        else
        {
            while(*ptr != '\n' && *ptr != '\r' && *ptr != '#' && *ptr != ' ' && *ptr != '\t')
                ptr++;
            *ptr = '\0';
            if (condition == 0)
                printf("Error: bad %s parameter value - %s, must be a non-negative integer\n", parameter, valptr);
            else if (condition == 1)
                printf("Error: bad %s parameter value - %s, must be a positive integer\n", parameter, valptr);
            else
                printf("Error: bad %s parameter value - %s, must be a power of 2 integer\n", parameter, valptr);
        }
        return 1;
    }
    
    *value = atoi(valptr);

    if (condition == 0)
        return 0;
    
    // check if it is zero
    if (*value == 0)
    {
        printf("Error: bad %s parameter value - %i, must be greater than 0\n", parameter, *value);
        return 1;
    }

    if (condition == 1)
        return 0;
    
    // check if it is power of 2
    temp = (int)(log(*value)/log(2));
    temp = (int)pow(2, temp);
    if (temp != *value)
    {
        printf("Error: bad %s parameter value - %i, must be power of 2\n", parameter, *value);
        return 1;
    }

    return 0;
}


int main(int argc, char *argv[])
{
    int ramsize, ramwidth, ramsizewidth;
    int Nmax, Nwidth, mdivk, m, mwidth, n;
    int ddrwidth, ddrsizewidth, memconwidth, ddrrowwidth, full_rate;
    int ddrburstlen,  burstlen, datawidth_ratio;
    int numPE, precision, PEwidth;
    int topsize, topwidth, topsizewidth;
    int ramlat, toplat;
    int multlat, addlat, divlat, totaldivlat;
    int intdivlat;
    int topinputdelay, topoutputdelay;
    int meminputdelay, memoutputdelay;
    int datawidth, ratio;
    int fifosize;
    int N, memaddr;
    int ddrDQperDQS, ddrrefclktick, compclktick;
    int diffDQS;

    FILE *fp;
    int ret;
    char str[500], parameter[500], name[500];
    char addname[500], multname[500], divname[500], memcname[500];
    char dirname[500];
    char *ptr1, *ptr2, *valptr, *fret;
    int value, temp, len;
    unsigned int coreParaDef, advParaDef, testParaDef;

    // core parameters
    /*numPE = 20;	        // number of PEs (mult + add)
    //ramsize = 128;	      // address size of on chip ram block
    mdivk = 2;          // internal block size in multiple of # of PEs
    ddrwidth = 24;      // ddr datawidth
    ddrsizewidth = 24;  // ddr total address width
    ddrrowwidth = 13;   // ddr row address width
    ddrburstlen = 4;    // ddr burst length
    ddrDQperDQS = 8;
    addlat = 13;			  // output latency of fp_add
    multlat = 10;			  // output latency of fp_mult
    divlat = 45;			  // output latency of fp_div
    precision = 32;     // bits of precision
    diffDQS = 0;
    ddrDQperDQS = 8;*/
    coreParaDef = 0x0fff;

    // advanced parameters
    // default values
    Nmax = 1000000;     // max matrix dimension
    fifosize = 16;      // size of fifo buffer
    topinputdelay = 3;  // # of registers in top block input ports
    topoutputdelay = 1; // # of registers in top block input ports
    meminputdelay = 2;  // # of registers in current and left blocks input ports
    memoutputdelay = 1; // # of registers in current and left blocks input ports
    advParaDef = 0x07ff;

    // test parameters (currently not used)
    N = 100;
    memaddr = 100;
    ddrrefclktick = 10000;
    compclktick = 5000;
    testParaDef = 0x000f;

    if (argc < 2)
    {
        printf("Error: missing paramter file name \nUsage - %s filename\n", argv[0]);
        return 1;
    }

    // open parameter file
    fp = fopen(argv[1], "r");
    if (fp == NULL)
    {
        printf("Error: parameter file \"%s\" cannot be opened\n", argv[1]);
        return 1;
    }

    // read from file
    fret = fgets(str, 500, fp);
    while (fret != NULL)
    {
        // find parameter name
        ptr1 = strchr(str, '=');
        ptr2 = strchr(str, '#');
        if (ptr1 == NULL || (ptr2 != NULL && ptr2 < ptr1))
        {
            // check if comment line
            ptr2=str;
            while(*ptr2 == ' ' || *ptr2 == '\t')
                ptr2++;
            if (*ptr2 != '#' && *ptr2 != '\n' && *ptr2 != '\r')
            {
                printf("Error: invalid format - %s", str);
                fclose(fp);
                return 1;
            }
            fret = fgets(str, 500, fp);
            continue;
        }
        valptr = ptr1+1;

        // get parameter name
        ptr2 = str;
        while(*ptr2 == ' ' || *ptr2 == '\t')
            ptr2++;
        ptr1--;
        while(*ptr1 == ' ' || *ptr1 == '\t')
            ptr1--;
        if (ptr1 < ptr2)
        {
            printf("Error: invalid format - %s", str);
            fclose(fp);
            return 1;
        }
        strncpy(parameter, ptr2, ptr1-ptr2+1);
        parameter[ptr1-ptr2+1] = '\0';

        // check if parameter and value is valid before assigning the value
        while(*valptr == ' ' || *valptr == '\t')
            valptr++;
        if(!strcmp(parameter,"name"))
        {
            ptr2 = valptr;
            while(*ptr2 != ' ' && *ptr2 != '\t' && *ptr2 != '\n' && *ptr2 != '\r' && *ptr2 != '#')
                ptr2++;
            strncpy(name, valptr, ptr2-valptr);
            name[ptr2-valptr] = '\0';
            if ((advParaDef & 0x0040) == 0)
                printf("Warning: %s parameter already assigned value, will overwrite previous assignment\n", parameter);
            advParaDef &= 0xffbf;
            printf("name is %s\n", name);
        }
        else if(!strcmp(parameter,"addname"))
        {
            ptr2 = valptr;
            while(*ptr2 != ' ' && *ptr2 != '\t' && *ptr2 != '\n' && *ptr2 != '\r' && *ptr2 != '#')
                ptr2++;
            strncpy(addname, valptr, ptr2-valptr);
            addname[ptr2-valptr] = '\0';
            if ((advParaDef & 0x0080) == 0)
                printf("Warning: %s parameter already assigned value, will overwrite previous assignment\n", parameter);
            advParaDef &= 0xff7f;
            printf("subname is %s\n", addname);
        }
        else if(!strcmp(parameter,"multname"))
        {
            ptr2 = valptr;
            while(*ptr2 != ' ' && *ptr2 != '\t' && *ptr2 != '\n' && *ptr2 != '\r' && *ptr2 != '#')
                ptr2++;
            if (ptr2-valptr == 0)
            {
                printf("Error: there is no %s parameter value\n");
                fclose(fp);
                return 1;
            }
            strncpy(multname, valptr, ptr2-valptr);
            multname[ptr2-valptr] = '\0';
            if ((advParaDef & 0x0100) == 0)
                printf("Warning: %s parameter already assigned value, will overwrite previous assignment\n", parameter);
            advParaDef &= 0xfeff;
            printf("multname is %s\n", multname);
        }
        else if(!strcmp(parameter,"divname"))
        {
            ptr2 = valptr;
            while(*ptr2 != ' ' && *ptr2 != '\t' && *ptr2 != '\n' && *ptr2 != '\r' && *ptr2 != '#')
                ptr2++;
            strncpy(divname, valptr, ptr2-valptr);
            divname[ptr2-valptr] = '\0';
            if ((advParaDef & 0x0200) == 0)
                printf("Warning: %s parameter already assigned value, will overwrite previous assignment\n", parameter);
            advParaDef &= 0xfdff;
            printf("divname is %s\n", divname);
        }
        else if(!strcmp(parameter,"memcontname"))
        {
            ptr2 = valptr;
            while(*ptr2 != ' ' && *ptr2 != '\t' && *ptr2 != '\n' && *ptr2 != '\r' && *ptr2 != '#')
                ptr2++;
            strncpy(memcname, valptr, ptr2-valptr);
            memcname[ptr2-valptr] = '\0';
            if ((advParaDef & 0x0400) == 0)
                printf("Warning: %s parameter already assigned value, will overwrite previous assignment\n", parameter);
            advParaDef &= 0xfbff;
            printf("memcontname is %s\n", memcname);
        }
        else 
        {
            if (!strcmp(parameter,"k"))
            {
                ret = checkVal(parameter, valptr, &value, 1);
                if (ret == 1)
                {
                    fclose(fp);
                    return 1;
                }
                numPE = value;
                if ((coreParaDef & 0x0001) == 0)
                    printf("Warning: %s parameter already assigned value, will overwrite previous assignment\n", parameter);
                coreParaDef &= 0xfffe;
            }
            else if (!strcmp(parameter,"mdivk"))
            {
                ret = checkVal(parameter, valptr, &value, 1);
                if (ret == 1)
                {
                    fclose(fp);
                    return 1;
                }
                mdivk = value;
                if ((coreParaDef & 0x0002) == 0)
                    printf("Warning: %s parameter already assigned value, will overwrite previous assignment\n", parameter);
                coreParaDef &= 0xfffd;
            }
            else if (!strcmp(parameter,"ddrwidth"))
            {
                ret = checkVal(parameter, valptr, &value, 1);
                if (ret == 1)
                {
                    fclose(fp);
                    return 1;
                }
                ddrwidth = value;
                if ((coreParaDef & 0x0004) == 0)
                    printf("Warning: %s parameter already assigned value, will overwrite previous assignment\n", parameter);
                coreParaDef &= 0xfffb;
            }
            else if (!strcmp(parameter,"ddrsizewidth"))
            {
                ret = checkVal(parameter, valptr, &value, 1);
                if (ret == 1)
                {
                    fclose(fp);
                    return 1;
                }
                ddrsizewidth = value;
                if ((coreParaDef & 0x0008) == 0)
                    printf("Warning: %s parameter already assigned value, will overwrite previous assignment\n", parameter);
                coreParaDef &= 0xfff7;
            }
            else if (!strcmp(parameter,"ddrrowwidth"))
            {
                ret = checkVal(parameter, valptr, &value, 1);
                if (ret == 1)
                {
                    fclose(fp);
                    return 1;
                }
                ddrrowwidth = value;
                if ((coreParaDef & 0x0010) == 0)
                    printf("Warning: %s parameter already assigned value, will overwrite previous assignment\n", parameter);
                coreParaDef &= 0xffef;
            }
            else if (!strcmp(parameter,"ddrburstlen"))
            {
                ret = checkVal(parameter, valptr, &value, 2);
                if (ret == 1)
                {
                    fclose(fp);
                    return 1;
                }
                ddrburstlen = value;
                if ((coreParaDef & 0x0020) == 0)
                    printf("Warning: %s parameter already assigned value, will overwrite previous assignment\n", parameter);
                coreParaDef &= 0xffdf;
            }
            else if (!strcmp(parameter,"addlat"))
            {
                ret = checkVal(parameter, valptr, &value, 1);
                if (ret == 1)
                {
                    fclose(fp);
                    return 1;
                }
                addlat = value;
                if ((coreParaDef & 0x0040) == 0)
                    printf("Warning: %s parameter already assigned value, will overwrite previous assignment\n", parameter);
                coreParaDef &= 0xffbf;
            }
            else if (!strcmp(parameter,"multlat"))
            {
                ret = checkVal(parameter, valptr, &value, 1);
                if (ret == 1)
                {
                    fclose(fp);
                    return 1;
                }
                multlat = value;
                if ((coreParaDef & 0x0080) == 0)
                    printf("Warning: %s parameter already assigned value, will overwrite previous assignment\n", parameter);
                coreParaDef &= 0xff7f;
            }
            else if (!strcmp(parameter,"divlat"))
            {
                ret = checkVal(parameter, valptr, &value, 1);
                if (ret == 1)
                {
                    fclose(fp);
                    return 1;
                }
                divlat = value;
                if ((coreParaDef & 0x0100) == 0)
                    printf("Warning: %s parameter already assigned value, will overwrite previous assignment\n", parameter);
                coreParaDef &= 0xfeff;
            }
            else if (!strcmp(parameter,"precision"))
            {
                ret = checkVal(parameter, valptr, &value, 0);
                if (ret == 1)
                {
                    fclose(fp);
                    return 1;
                }
                if (value != 32 && value != 64)
                {
                    printf("Error: bad %s parameter value - %i, must be either 32 or 64\n", parameter, value);
                    fclose(fp);
                    return 1;
                }
                precision = value;
                if ((coreParaDef & 0x0200) == 0)
                    printf("Warning: %s parameter already assigned value, will overwrite previous assignment\n", parameter);
                coreParaDef &= 0xfdff;
            }
            else if (!strcmp(parameter,"DQperDQS"))
            {
                ret = checkVal(parameter, valptr, &value, 0);
                if (ret == 1)
                {
                    fclose(fp);
                    return 1;
                }
                if (value != 4 && value != 8)
                {
                    printf("Error: bad %s parameter value - %i, must be either 4 or 8\n", parameter, value);
                    fclose(fp);
                    return 1;
                }
                ddrDQperDQS = value;
                if ((coreParaDef & 0x0400) == 0)
                    printf("Warning: %s parameter already assigned value, will overwrite previous assignment\n", parameter);
                coreParaDef &= 0xfbff;
            }
            else if (!strcmp(parameter,"diffDQS"))
            {
                ret = checkVal(parameter, valptr, &value, 0);
                if (ret == 1)
                {
                    fclose(fp);
                    return 1;
                }
                value = atoi(valptr);
                if (value != 0 && value != 1)
                {
                    printf("Error: bad %s parameter value - %i, must be either 0 or 1\n", parameter, value);
                    fclose(fp);
                    return 1;
                }
                diffDQS = value;
                if ((coreParaDef & 0x0800) == 0)
                    printf("Warning: %s parameter already assigned value, will overwrite previous assignment\n", parameter);
                coreParaDef &= 0xf7ff;
            }
            else if (!strcmp(parameter,"Nmax"))
            {
                ret = checkVal(parameter, valptr, &value, 1);
                if (ret == 1)
                {
                    fclose(fp);
                    return 1;
                }
                Nmax = value;
                if ((advParaDef & 0x0001) == 0)
                    printf("Warning: %s parameter already assigned value, will overwrite previous assignment\n", parameter);
                advParaDef &= 0xfffe;
            }
            else if (!strcmp(parameter,"fifosize"))
            {
                ret = checkVal(parameter, valptr, &value, 2);
                if (ret == 1)
                {
                    fclose(fp);
                    return 1;
                }
                fifosize = value;
                if ((advParaDef & 0x0002) == 0)
                    printf("Warning: %s parameter already assigned value, will overwrite previous assignment\n", parameter);
                advParaDef &= 0xfffd;
            }
            else if (!strcmp(parameter,"meminputdelay"))
            {
                ret = checkVal(parameter, valptr, &value, 0);
                if (ret == 1)
                {
                    fclose(fp);
                    return 1;
                }
                meminputdelay = value;
                if ((advParaDef & 0x0004) == 0)
                    printf("Warning: %s parameter already assigned value, will overwrite previous assignment\n", parameter);
                advParaDef &= 0xfffb;
            }
            else if (!strcmp(parameter,"memoutputdelay"))
            {
                ret = checkVal(parameter, valptr, &value, 0);
                if (ret == 1)
                {
                    fclose(fp);
                    return 1;
                }
                memoutputdelay = value;
                if ((advParaDef & 0x0008) == 0)
                    printf("Warning: %s parameter already assigned value, will overwrite previous assignment\n", parameter);
                advParaDef &= 0xfff7;
            }
            else if (!strcmp(parameter,"topinputdelay"))
            {
                ret = checkVal(parameter, valptr, &value, 0);
                if (ret == 1)
                {
                    fclose(fp);
                    return 1;
                }
                topinputdelay = value;
                if ((advParaDef & 0x0010) == 0)
                    printf("Warning: %s parameter already assigned value, will overwrite previous assignment\n", parameter);
                advParaDef &= 0xffef;
            }
            else if (!strcmp(parameter,"topoutputdelay"))
            {
                ret = checkVal(parameter, valptr, &value, 0);
                if (ret == 1)
                {
                    fclose(fp);
                    return 1;
                }
                topoutputdelay = value;
                if ((advParaDef & 0x0020) == 0)
                    printf("Warning: %s parameter already assigned value, will overwrite previous assignment\n", parameter);
                advParaDef &= 0xffdf;
            }
            else if (!strcmp(parameter, "N"))
            {
                ret = checkVal(parameter, valptr, &value, 1);
                if (ret == 1)
                {
                    fclose(fp);
                    return 1;
                }
                N = value;
                testParaDef &= 0xfffe;
            }
            else if (!strcmp(parameter, "startmemaddr"))
            {
                ret = checkVal(parameter, valptr, &value, 0);
                if (ret == 1)
                {
                    fclose(fp);
                    return 1;
                }
                memaddr = value;
                testParaDef &= 0xfffd;
            }
            else if (!strcmp(parameter, "ddrrefclktick"))
            {
                ret = checkVal(parameter, valptr, &value, 1);
                if (ret == 1)
                {
                    fclose(fp);
                    return 1;
                }
                ddrrefclktick = value;
                testParaDef &= 0xfffb;
            }
            else if (!strcmp(parameter, "compclktick"))
            {
                ret = checkVal(parameter, valptr, &value, 1);
                if (ret == 1)
                {
                    fclose(fp);
                    return 1;
                }
                compclktick = value;
                testParaDef &= 0xfff7;
            }
            else
            {
                printf("Error:unknown parameter name - %s\n", parameter);
                return 1;
            }
            printf("%s is set to %i\n", parameter, value);
        }
        fret = fgets(str, 500, fp);
    }
    fclose(fp);

    // check if any core parameters are not specified
    if (coreParaDef != 0)
    {
        printf("Error: the following missing core parameters\n");
        if ((coreParaDef & 0x0001) != 0)
            printf("k\n");
        if ((coreParaDef & 0x0002) != 0)
            printf("mdivk\n");
        if ((coreParaDef & 0x0004) != 0)
            printf("ddrwidth\n");
        if ((coreParaDef & 0x0008) != 0)
            printf("ddrsizewidth\n");
        if ((coreParaDef & 0x0010) != 0)
            printf("ddrrowwidth\n");
        if ((coreParaDef & 0x0020) != 0)
            printf("ddrburstlen\n");
        if ((coreParaDef & 0x0040) != 0)
            printf("addlat\n");
        if ((coreParaDef & 0x0080) != 0)
            printf("multlat\n");
        if ((coreParaDef & 0x0100) != 0)
            printf("divlat\n");
        if ((coreParaDef & 0x0200) != 0)
            printf("precision\n");
        if ((coreParaDef & 0x0400) != 0)
            printf("ddrDQperDQS\n");
        if ((coreParaDef & 0x0800) != 0)
            printf("diffDQS\n");
        return 1;
    }


    // print out default parameter values if there are some defaulted parameters
    if ((advParaDef & 0x0001) != 0)
        printf("Nmax is defaulted to %i\n", Nmax);
    if ((advParaDef & 0x0002) != 0)
        printf("fifosize is defaulted to %i\n", fifosize);
    if ((advParaDef & 0x0004) != 0)
        printf("meminputdelay is defaulted to %i\n", meminputdelay);
    if ((advParaDef & 0x0008) != 0)
        printf("memoutputdelay is defaulted to %i\n", memoutputdelay);
    if ((advParaDef & 0x0010) != 0)
        printf("topinputdelay is defaulted to %i\n", topinputdelay);
    if ((advParaDef & 0x0020) != 0)
        printf("topoutputdelay is defaulted to %i\n", topoutputdelay);
    if ((advParaDef & 0x0040) != 0)
    {
        sprintf(name, "LU%iPEEng", numPE);
        printf("name is defaulted to %s\n", name);
    }
    if ((advParaDef & 0x0080) != 0)
    {
        if (precision == 64)
            sprintf(addname, "subdplat%i", addlat);
        else
            sprintf(addname, "subsplat%i", addlat);
        printf("addname is defaulted to %s\n", addname);
    }
    if ((advParaDef & 0x0100) != 0)
    {
        if (precision == 64)
            sprintf(multname, "multdplat%i", multlat);
        else
            sprintf(multname, "multsplat%i", multlat);
        printf("multname is defaulted to %s\n", multname);
    }
    if ((advParaDef & 0x0200) != 0)
    {
        if (precision == 64)
            sprintf(divname, "divdp");
        else
            sprintf(divname, "divsp");
        printf("divname is defaulted to %s\n", divname);
    }
    if ((advParaDef & 0x0400) != 0)
    {
        sprintf(memcname, "ddr2_%ix%i_%i", ddrsizewidth, ddrwidth, ddrDQperDQS);
        printf("memcontname is defaulted to %s\n", memcname);
    }

    //test parameter
    /*
    if ((testParaDef & 0x0001) != 0)
        printf("N is defaulted to %i\n", N);
    if ((testParaDef & 0x0002) != 0)
        printf("startmemaddr is defaulted to %i\n", memaddr);
    if ((testParaDef & 0x0004) != 0)
        printf("ddrrefclktick is defaulted to %i\n", ddrrefclktick);
    if ((testParaDef & 0x0008) != 0)
        printf("compclktick is defaulted to %i\n", compclktick);
    */
    
    ret = chdir(name);
    //printf("ret for chdir : %i\n", ret);
    if (ret != 0)
    {
        mkdir(name, S_IRWXU);
        ret = chdir(name);
        if (ret != 0)
        {
            printf("Error: failed to create directory with name \"%s\"\n", name);
            return 1;
        }
        //printf("ret for chdir : %i\n", ret);
    }

    //fp = fopen("testing.txt","w");
    //fprintf(fp, "success");
    //fclose(fp);

    totaldivlat = divlat;
    ramlat = 2 + meminputdelay + memoutputdelay;			// total latency for reading ram block
    toplat = 2 + topinputdelay + topoutputdelay;			// total latency for reading temp_row block
    intdivlat = 15;
    full_rate = 1;
    m = numPE*mdivk;
    n = m;
    ramwidth = numPE*precision;
    //ramsize = numPE*mdivk*mdivk;
    //topsize = numPE*mdivk*mdivk*numPE; //ramsize*numPE;   // address size of temp_row block
    datawidth_ratio = (full_rate == 0) ? 4 : 2;
    burstlen = ddrburstlen/datawidth_ratio;    
    memconwidth = ddrwidth*datawidth_ratio;
    ramwidth = numPE*precision;
    topwidth = precision;
    ramsizewidth = intlog2(numPE*mdivk*mdivk);
    topsizewidth = intlog2(numPE*mdivk*mdivk*numPE);
    ramsize = pow(2, ramsizewidth);
    topsize = pow(2, topsizewidth);
    Nwidth = intlog2(Nmax+1);
    mwidth = intlog2(m+2);
    ratio = pow(2, ceil(log(ramwidth*1.0/memconwidth)/log(2)));
    datawidth = memconwidth*ratio;

    // call functions to produce modules in compute engine
    
    gentop(name, memcname, Nwidth, mwidth, ddrwidth, ddrDQperDQS, ddrsizewidth, burstlen, ddrrowwidth, memconwidth, ramwidth, ramsizewidth, diffDQS);

    genBlockLU(divname, numPE, precision, mwidth, ramwidth, ramsizewidth, topsizewidth,
        meminputdelay, memoutputdelay, topinputdelay, topoutputdelay);

    genBlockControl(numPE, precision, mwidth, mdivk, intdivlat, multlat, addlat,
        ramlat, ramlat, toplat, totaldivlat, ramwidth, ramsizewidth, topsizewidth);

    genMC(m, Nwidth, mdivk, ratio, ddrsizewidth, ramsizewidth);

    genDTU(burstlen, datawidth, memconwidth, ddrsizewidth, fifosize,
        ramwidth, ramsizewidth, ratio, ramlat, mwidth);

    genram(ramsize, ramwidth);

    gentop_ram(topsize, topwidth);

    genmult_add(addname, multname, precision);

    genmemcmd_fifo(fifosize, burstlen, ddrsizewidth);

    genaddr_fifo(fifosize, ramsizewidth);

    genwfifo(fifosize, memconwidth, ratio);

    genrfifo(fifosize, memconwidth, ratio);

    //genLU_testbench(name, N, memaddr, ddrrowwidth, ddrDQperDQS, datawidth_ratio, memconwidth, ddrrefclktick, compclktick);

    //genmem(N, numPE, memconwidth, ratio, mdivk, memaddr);

    gensdc(name, divname, compclktick, intdivlat);


    // copy core files
    chdir("..");
    len = strlen(corelibname);
    strcpy(dirname, corelibname);
    dirname[len] = '/';
    
    strcpy(dirname+len+1, addname);
    if (copydir(dirname, name))
        return 1;

    strcpy(dirname+len+1, multname);
    if (copydir(dirname, name))
        return 1;

    strcpy(dirname+len+1, divname);
    if (copydir(dirname, name))
        return 1;

    strcpy(dirname+len+1, memcname);
    if (copydir(dirname, name))
        return 1;

	return 0;
}
