// generate addr_fifo module
// by Wei Zhang
//

#include <stdio.h>
#include <math.h>
#define intlog2(x) (int)ceil(log(x)/log(2))

int genaddr_fifo(int fifosize, int ramsizewidth)
{
	FILE *fp;
  int fifosizewidth, fifowidth;

  fifosizewidth = intlog2(fifosize);
  fifowidth = ramsizewidth;

	fp = fopen("addr_fifo.v", "w");

	fprintf(fp, "// synopsys translate_off\n");
	fprintf(fp, "`timescale 1 ps / 1 ps\n");
	fprintf(fp, "// synopsys translate_on\n");
	fprintf(fp, "module addr_fifo (\n");
	fprintf(fp, "	clock,\n");
	fprintf(fp, "	data,\n");
	fprintf(fp, "	rdreq,\n");
	fprintf(fp, "	wrreq,\n");
	fprintf(fp, "	empty,\n");
	fprintf(fp, "	full,\n");
	fprintf(fp, "	q);\n");
	fprintf(fp, "\n");
	fprintf(fp, "	parameter FIFOSIZE = %i, FIFOSIZEWIDTH = %i, FIFOWIDTH = %i;\n", fifosize, fifosizewidth, fifowidth);
	fprintf(fp, "\n");
	fprintf(fp, "	input	  clock;\n");
	fprintf(fp, "	input	[FIFOWIDTH-1:0]  data;\n");
	fprintf(fp, "	input	  rdreq;\n");
	fprintf(fp, "	input	  wrreq;\n");
	fprintf(fp, "	output	  empty;\n");
	fprintf(fp, "	output	  full;\n");
	fprintf(fp, "	output	[FIFOWIDTH-1:0]  q;\n");
	fprintf(fp, "\n");
	fprintf(fp, "	wire  sub_wire0;\n");
	fprintf(fp, "	wire [FIFOWIDTH-1:0] sub_wire1;\n");
	fprintf(fp, "	wire  sub_wire2;\n");
	fprintf(fp, "	wire  empty = sub_wire0;\n");
	fprintf(fp, "	wire [FIFOWIDTH-1:0] q = sub_wire1[FIFOWIDTH-1:0];\n");
	fprintf(fp, "	wire  full = sub_wire2;\n");
	fprintf(fp, "\n");
	fprintf(fp, "	scfifo	scfifo_component (\n");
	fprintf(fp, "				.rdreq (rdreq),\n");
	fprintf(fp, "				.clock (clock),\n");
	fprintf(fp, "				.wrreq (wrreq),\n");
	fprintf(fp, "				.data (data),\n");
	fprintf(fp, "				.empty (sub_wire0),\n");
	fprintf(fp, "				.q (sub_wire1),\n");
	fprintf(fp, "				.full (sub_wire2)\n");
	fprintf(fp, "				// synopsys translate_off\n");
	fprintf(fp, "				,\n");
	fprintf(fp, "				.aclr (),\n");
	fprintf(fp, "				.almost_empty (),\n");
	fprintf(fp, "				.almost_full (),\n");
	fprintf(fp, "				.sclr (),\n");
	fprintf(fp, "				.usedw ()\n");
	fprintf(fp, "				// synopsys translate_on\n");
	fprintf(fp, "				);\n");
	fprintf(fp, "	defparam\n");
	fprintf(fp, "		scfifo_component.add_ram_output_register = \"ON\",\n");
	fprintf(fp, "		scfifo_component.intended_device_family = \"Stratix III\",\n");
	fprintf(fp, "		scfifo_component.lpm_numwords = FIFOSIZE,\n");
	fprintf(fp, "		scfifo_component.lpm_showahead = \"OFF\",\n");
	fprintf(fp, "		scfifo_component.lpm_type = \"scfifo\",\n");
	fprintf(fp, "		scfifo_component.lpm_width = FIFOWIDTH,\n");
	fprintf(fp, "		scfifo_component.lpm_widthu = FIFOSIZEWIDTH,\n");
	fprintf(fp, "		scfifo_component.overflow_checking = \"OFF\",\n");
	fprintf(fp, "		scfifo_component.underflow_checking = \"OFF\",\n");
	fprintf(fp, "		scfifo_component.use_eab = \"ON\";\n");
	fprintf(fp, "\n");
	fprintf(fp, "\n");
	fprintf(fp, "endmodule\n");

	return 0;
}
