// generate memcmd_fifo module
// by Wei Zhang
//

#include <stdio.h>
#include <math.h>
#define intlog2(x) (int)ceil(log(x)/log(2))

int genmemcmd_fifo(int fifosize, int burstlen, int ddrsizewidth)
{
	FILE *fp;
  int burstwidth, fifosizewidth, fifowidth;

  burstwidth = intlog2(burstlen+1);
  fifosizewidth = intlog2(fifosize);
  fifowidth = ddrsizewidth + burstwidth + 2;

	fp = fopen("memcmd_fifo.v", "w");

	fprintf(fp, "// synopsys translate_off\n");
	fprintf(fp, "`timescale 1 ps / 1 ps\n");
	fprintf(fp, "// synopsys translate_on\n");
	fprintf(fp, "module memcmd_fifo (\n");
	fprintf(fp, "	aclr,\n");
  fprintf(fp, "	data,\n");
	fprintf(fp, "	rdclk,\n");
	fprintf(fp, "	rdreq,\n");
	fprintf(fp, "	wrclk,\n");
	fprintf(fp, "	wrreq,\n");
	fprintf(fp, "	q,\n");
	fprintf(fp, "	rdempty,\n");
	fprintf(fp, "	wrempty,\n");
	fprintf(fp, "	wrfull);\n");
	fprintf(fp, "	\n");
	fprintf(fp, "	parameter FIFOSIZE = %i, FIFOSIZEWIDTH = %i, FIFOWIDTH = %i;\n", fifosize, fifosizewidth, fifowidth);
	fprintf(fp, "\n");
	fprintf(fp, "	input	  aclr;\n");
  fprintf(fp, "	input	[FIFOWIDTH-1:0]  data;\n");
	fprintf(fp, "	input	  rdclk;\n");
	fprintf(fp, "	input	  rdreq;\n");
	fprintf(fp, "	input	  wrclk;\n");
	fprintf(fp, "	input	  wrreq;\n");
	fprintf(fp, "	output	[FIFOWIDTH-1:0]  q;\n");
	fprintf(fp, "	output	  rdempty;\n");
	fprintf(fp, "	output	  wrempty;\n");
	fprintf(fp, "	output	  wrfull;\n");
	fprintf(fp, "\n");
	fprintf(fp, "	wire  sub_wire0;\n");
	fprintf(fp, "	wire  sub_wire1;\n");
	fprintf(fp, "	wire  sub_wire2;\n");
	fprintf(fp, "	wire [FIFOWIDTH-1:0] sub_wire3;\n");
	fprintf(fp, "	wire  rdempty = sub_wire0;\n");
	fprintf(fp, "	wire  wrfull = sub_wire1;\n");
	fprintf(fp, "	wire  wrempty = sub_wire2;\n");
	fprintf(fp, "	wire [FIFOWIDTH-1:0] q = sub_wire3[FIFOWIDTH-1:0];\n");
	fprintf(fp, "\n");
	fprintf(fp, "	dcfifo	dcfifo_component (\n");
	fprintf(fp, "				.wrclk (wrclk),\n");
	fprintf(fp, "				.rdreq (rdreq),\n");
 	fprintf(fp, "				.aclr (alcr),\n");
	fprintf(fp, "				.rdclk (rdclk),\n");
	fprintf(fp, "				.wrreq (wrreq),\n");
	fprintf(fp, "				.data (data),\n");
	fprintf(fp, "				.rdempty (sub_wire0),\n");
	fprintf(fp, "				.wrfull (sub_wire1),\n");
	fprintf(fp, "				.wrempty (sub_wire2),\n");
	fprintf(fp, "				.q (sub_wire3)\n");
	fprintf(fp, "				// synopsys translate_off\n");
	fprintf(fp, "				,\n");
	fprintf(fp, "				.rdfull (),\n");
	fprintf(fp, "				.rdusedw (),\n");
	fprintf(fp, "				.wrusedw ()\n");
	fprintf(fp, "				// synopsys translate_on\n");
	fprintf(fp, "				);\n");
	fprintf(fp, "	defparam\n");
	fprintf(fp, "		dcfifo_component.intended_device_family = \"Stratix III\",\n");
	fprintf(fp, "		dcfifo_component.lpm_numwords = FIFOSIZE,\n");
	fprintf(fp, "		dcfifo_component.lpm_showahead = \"OFF\",\n");
	fprintf(fp, "		dcfifo_component.lpm_type = \"dcfifo\",\n");
	fprintf(fp, "		dcfifo_component.lpm_width = FIFOWIDTH,\n");
	fprintf(fp, "		dcfifo_component.lpm_widthu = FIFOSIZEWIDTH,\n");
	fprintf(fp, "		dcfifo_component.overflow_checking = \"OFF\",\n");
	fprintf(fp, "		dcfifo_component.rdsync_delaypipe = 5,\n");
	fprintf(fp, "		dcfifo_component.underflow_checking = \"OFF\",\n");
	fprintf(fp, "		dcfifo_component.use_eab = \"ON\",\n");
	fprintf(fp, "		dcfifo_component.write_aclr_synch = \"OFF\",\n");
	fprintf(fp, "		dcfifo_component.wrsync_delaypipe = 5;\n");
	fprintf(fp, "\n");
	fprintf(fp, "\n");
	fprintf(fp, "endmodule\n");

	return 0;
}
