// generate ram module
// by Wei Zhang
//

#include <stdio.h>
#include <math.h>
#define intlog2(x) (int)ceil(log(x)/log(2))

int genram(int ramsize, int ramwidth)
{
	FILE *fp;
  int ramsizewidth;

  ramsizewidth = intlog2(ramsize);

	fp = fopen("ram.v", "w");

	fprintf(fp, "// synopsys translate_off\n");
	fprintf(fp, "`timescale 1 ps / 1 ps\n");
	fprintf(fp, "// synopsys translate_on\n");
	fprintf(fp, "module ram (\n");
	fprintf(fp, "	byteena_a,\n");
	fprintf(fp, "	clock,\n");
	fprintf(fp, "	data,\n");
	fprintf(fp, "	rdaddress,\n");
	fprintf(fp, "	wraddress,\n");
	fprintf(fp, "	wren,\n");
	fprintf(fp, "	q);\n");
	fprintf(fp, "\n");
	fprintf(fp, "	parameter RAMWIDTH = %i, RAMNUMBYTES = %i, RAMSIZE = %i, RAMSIZEWIDTH = %i;\n", ramwidth, ramwidth/8, ramsize, ramsizewidth);
	fprintf(fp, "\n");
	fprintf(fp, "	input	[RAMNUMBYTES-1:0]  byteena_a;\n");
	fprintf(fp, "	input	  clock;\n");
	fprintf(fp, "	input	[RAMWIDTH-1:0]  data;\n");
	fprintf(fp, "	input	[RAMSIZEWIDTH-1:0]  rdaddress;\n");
	fprintf(fp, "	input	[RAMSIZEWIDTH-1:0]  wraddress;\n");
	fprintf(fp, "	input	  wren;\n");
	fprintf(fp, "	output	[RAMWIDTH-1:0]  q;\n");
	fprintf(fp, "\n");
	fprintf(fp, "	wire [RAMWIDTH-1:0] sub_wire0;\n");
	fprintf(fp, "	wire [RAMWIDTH-1:0] q = sub_wire0[RAMWIDTH-1:0];\n");
	fprintf(fp, "\n");
	fprintf(fp, "	altsyncram	altsyncram_component (\n");
	fprintf(fp, "				.wren_a (wren),\n");
	fprintf(fp, "				.clock0 (clock),\n");
	fprintf(fp, "				.byteena_a (byteena_a),\n");
	fprintf(fp, "				.address_a (wraddress),\n");
	fprintf(fp, "				.address_b (rdaddress),\n");
	fprintf(fp, "				.data_a (data),\n");
	fprintf(fp, "				.q_b (sub_wire0),\n");
	fprintf(fp, "				.aclr0 (1'b0),\n");
	fprintf(fp, "				.aclr1 (1'b0),\n");
	fprintf(fp, "				.addressstall_a (1'b0),\n");
	fprintf(fp, "				.addressstall_b (1'b0),\n");
	fprintf(fp, "				.byteena_b (1'b1),\n");
	fprintf(fp, "				.clock1 (1'b1),\n");
	fprintf(fp, "				.clocken0 (1'b1),\n");
	fprintf(fp, "				.clocken1 (1'b1),\n");
	fprintf(fp, "				.clocken2 (1'b1),\n");
	fprintf(fp, "				.clocken3 (1'b1),\n");
	fprintf(fp, "				.data_b ({RAMWIDTH{1'b1}}),\n");
	fprintf(fp, "				.eccstatus (),\n");
	fprintf(fp, "				.q_a (),\n");
	fprintf(fp, "				.rden_a (1'b1),\n");
	fprintf(fp, "				.rden_b (1'b1),\n");
	fprintf(fp, "				.wren_b (1'b0));\n");
	fprintf(fp, "	defparam\n");
	fprintf(fp, "		altsyncram_component.address_aclr_b = \"NONE\",\n");
	fprintf(fp, "		altsyncram_component.address_reg_b = \"CLOCK0\",\n");
	fprintf(fp, "		altsyncram_component.byte_size = 8,\n");
	fprintf(fp, "		altsyncram_component.clock_enable_input_a = \"BYPASS\",\n");
	fprintf(fp, "		altsyncram_component.clock_enable_input_b = \"BYPASS\",\n");
	fprintf(fp, "		altsyncram_component.clock_enable_output_b = \"BYPASS\",\n");
	fprintf(fp, "		altsyncram_component.intended_device_family = \"Stratix III\",\n");
	fprintf(fp, "		altsyncram_component.lpm_type = \"altsyncram\",\n");
	fprintf(fp, "		altsyncram_component.numwords_a = RAMSIZE,\n");
	fprintf(fp, "		altsyncram_component.numwords_b = RAMSIZE,\n");
	fprintf(fp, "		altsyncram_component.operation_mode = \"DUAL_PORT\",\n");
	fprintf(fp, "		altsyncram_component.outdata_aclr_b = \"NONE\",\n");
	fprintf(fp, "		altsyncram_component.outdata_reg_b = \"CLOCK0\",\n");
	fprintf(fp, "		altsyncram_component.power_up_uninitialized = \"FALSE\",\n");
	fprintf(fp, "		altsyncram_component.read_during_write_mode_mixed_ports = \"DONT_CARE\",\n");
	fprintf(fp, "		altsyncram_component.widthad_a = RAMSIZEWIDTH,\n");
	fprintf(fp, "		altsyncram_component.widthad_b = RAMSIZEWIDTH,\n");
	fprintf(fp, "		altsyncram_component.width_a = RAMWIDTH,\n");
	fprintf(fp, "		altsyncram_component.width_b = RAMWIDTH,\n");
	fprintf(fp, "		altsyncram_component.width_byteena_a = RAMNUMBYTES;\n");
	fprintf(fp, "\n");
	fprintf(fp, "\n");
	fprintf(fp, "endmodule\n");

	return 0;
}
