// generate top module
// by Wei Zhang
//

#include <stdio.h>
#include <math.h>
#include <string.h>

#define intlog2(x) (int)ceil(log(x)/log(2))

void gentop(const char * name, const char *memcname, int Nwidth, int mwidth, int ddrwidth, int ddrDQperDQS, int ddrsizewidth, int burstlen, int ddrrowwidth,
    int memconwidth, int ramwidth, int ramsizewidth, int mem_dqsn)
{
    FILE *fp;
    int len;
    char filename[500];

    strcpy(filename, name);
    len = strlen(filename);
    filename[len] = '.';
    filename[len+1] = 'v';
    filename[len+2] = '\0';

    fp = fopen(filename, "w");

    fprintf(fp, "//auto-generated top.v\n");
    fprintf(fp, "//top level module of LU factorization\n");
    fprintf(fp, "//by Wei Zhang\n\n");
    fprintf(fp, "module %s (clk, ref_clk, global_reset_n, start, N, offset, done, ready,\n", name);
    fprintf(fp, "		mem_addr, mem_ba, mem_cas_n, mem_cke, mem_clk, mem_clk_n, mem_cs_n,\n");
    if (mem_dqsn != 0)
    {
        fprintf(fp, "		mem_dm, mem_dq, mem_dqs, mem_dqsn, mem_odt, mem_ras_n, mem_we_n);\n");
    }
    else
    {
        fprintf(fp, "		mem_dm, mem_dq, mem_dqs, mem_odt, mem_ras_n, mem_we_n);\n");
    }
    fprintf(fp, "\n");
    fprintf(fp, "parameter NWIDTH = %i, BLOCKWIDTH = %i;\n", Nwidth, mwidth);
    fprintf(fp, "parameter DDRWIDTH = %i, DDRNUMDQS = %i, DDRSIZEWIDTH = %i;\n", ddrwidth, ddrwidth/ddrDQperDQS, ddrsizewidth);
    fprintf(fp, "parameter BURSTLEN = %i;\n", burstlen);
    fprintf(fp, "parameter MEMCONWIDTH = %i, MEMCONNUMBYTES = %i;\n", memconwidth, memconwidth/8);
    fprintf(fp, "parameter RAMWIDTH = %i, RAMNUMBYTES = %i, RAMSIZEWIDTH = %i;\n", ramwidth, ramwidth/8, ramsizewidth);
    fprintf(fp, "\n");
    fprintf(fp, "input start;\n");
    fprintf(fp, "input[NWIDTH-1:0] N;\n");
    fprintf(fp, "input[DDRSIZEWIDTH-1:0] offset;\n");
    fprintf(fp, "output done;\n");
    fprintf(fp, "output [%i:0] mem_addr;\n", ddrrowwidth-1);
    fprintf(fp, "output [1:0] mem_ba;\n");
    fprintf(fp, "output mem_cas_n;\n");
    fprintf(fp, "output [0:0] mem_cke;\n");
    fprintf(fp, "inout [0:0] mem_clk;\n");
    fprintf(fp, "inout [0:0] mem_clk_n;\n");
    fprintf(fp, "output [0:0] mem_cs_n;\n");
    fprintf(fp, "output [DDRNUMDQS-1:0] mem_dm;\n");
    fprintf(fp, "inout [DDRWIDTH-1:0] mem_dq;\n");
    fprintf(fp, "inout [DDRNUMDQS-1:0] mem_dqs;\n");
    if (mem_dqsn != 0)
    {
        fprintf(fp, "inout [DDRNUMDQS-1:0] mem_dqsn;\n");
    }
    fprintf(fp, "output [0:0] mem_odt;\n");
    fprintf(fp, "output mem_ras_n;\n");
    fprintf(fp, "output mem_we_n;\n");
    fprintf(fp, "input clk, ref_clk;\n");
    fprintf(fp, "input global_reset_n;\n");
    fprintf(fp, "output ready;\n");
    fprintf(fp, "\n");
    fprintf(fp, "wire phy_clk;\n");
    fprintf(fp, "wire[BLOCKWIDTH-1:0] m, n, loop;\n");
    fprintf(fp, "wire[1:0] mode;\n");
    fprintf(fp, "wire comp_start, comp_done;\n");
    fprintf(fp, "wire dtu_write_req, dtu_read_req, dtu_ack, dtu_done;\n");
    fprintf(fp, "wire [DDRSIZEWIDTH-1:0] dtu_mem_addr;\n");
    fprintf(fp, "wire [RAMSIZEWIDTH-1:0] dtu_ram_addr;\n");
    fprintf(fp, "wire [BLOCKWIDTH-1:0] dtu_size;\n");
    fprintf(fp, "wire left_sel;\n");
    fprintf(fp, "\n");
    fprintf(fp, "wire[RAMWIDTH-1:0] curWriteDataMem, curReadDataMem;\n");
    fprintf(fp, "wire[RAMSIZEWIDTH-1:0] curWriteAddrMem, curReadAddrMem;\n");
    fprintf(fp, "wire[RAMNUMBYTES-1:0] curWriteByteEnMem;\n");
    fprintf(fp, "wire curWriteEnMem;\n");
    fprintf(fp, "wire[RAMWIDTH-1:0] leftWriteDataMem;\n");
    fprintf(fp, "wire[RAMSIZEWIDTH-1:0] leftWriteAddrMem;\n");
    fprintf(fp, "wire[RAMNUMBYTES-1:0] leftWriteByteEnMem;\n");
    fprintf(fp, "wire leftWriteEnMem;\n");
    fprintf(fp, "wire curMemSel, leftMemSel;\n");
    fprintf(fp, "\n");
    fprintf(fp, "wire burst_begin;\n");
    fprintf(fp, "wire [MEMCONNUMBYTES-1:0] mem_local_be;\n");
    fprintf(fp, "wire mem_local_read_req;\n");
    fprintf(fp, "wire [BURSTLEN-1:0] mem_local_size;\n");
    fprintf(fp, "wire [MEMCONWIDTH-1:0] mem_local_wdata;\n");
    fprintf(fp, "wire mem_local_write_req;\n");
    fprintf(fp, "wire [MEMCONWIDTH-1:0] mem_local_rdata;\n");
    fprintf(fp, "wire mem_local_rdata_valid;\n");
    fprintf(fp, "wire mem_local_ready;\n");
    fprintf(fp, "wire mem_local_wdata_req;\n");
    fprintf(fp, "wire reset_n;\n");
    fprintf(fp, "wire [DDRSIZEWIDTH-1:0] mem_local_addr;\n");
    fprintf(fp, "\n");
    fprintf(fp, "wire[RAMWIDTH-1:0] ram_write_data, ram_read_data;\n");
    fprintf(fp, "wire[RAMSIZEWIDTH-1:0] ram_write_addr, ram_read_addr;\n");
    fprintf(fp, "wire[RAMNUMBYTES-1:0] ram_write_byte_en;\n");
    fprintf(fp, "wire ram_write_en;\n");
    fprintf(fp, "\n");
    fprintf(fp, "MarshallerController MC (clk, start, done, N, offset,\n");
    fprintf(fp, "	comp_start, m, n, loop, mode, comp_done, curMemSel, leftMemSel,\n");
    fprintf(fp, "	dtu_write_req, dtu_read_req, dtu_mem_addr, dtu_ram_addr, dtu_size, dtu_ack, dtu_done, left_sel);\n");
    fprintf(fp, "\n");
    fprintf(fp, "// block that computes the LU factorization, with answer stored back into ram block\n");
    fprintf(fp, "LU compBlock (clk, comp_start, m, n, loop, mode, comp_done,\n");
    fprintf(fp, "			curReadAddrMem, curReadDataMem, curWriteByteEnMem, curWriteDataMem, curWriteAddrMem, curWriteEnMem, curMemSel,\n");
    fprintf(fp, "			leftWriteByteEnMem, leftWriteDataMem, leftWriteAddrMem, leftWriteEnMem, leftMemSel);\n");
    fprintf(fp, "\n");
    fprintf(fp, "DataTransferUnit DTU (clk, phy_clk, dtu_write_req, dtu_read_req, dtu_mem_addr, dtu_ram_addr, dtu_size, dtu_ack, dtu_done,\n");
    fprintf(fp, "		ram_read_addr, ram_read_data, ram_write_byte_en, ram_write_data, ram_write_addr, ram_write_en,\n");
    fprintf(fp, "		mem_local_rdata, mem_local_rdata_valid, mem_local_ready, mem_local_wdata_req, reset_n,\n");
    fprintf(fp, "		burst_begin, mem_local_addr, mem_local_be, mem_local_read_req, mem_local_size,\n");
    fprintf(fp, "		mem_local_wdata, mem_local_write_req);\n");
    fprintf(fp, "\n");
    fprintf(fp, "%s memController (\n", memcname);
    fprintf(fp, "	.global_reset_n (global_reset_n),\n");
    fprintf(fp, "	.local_address (mem_local_addr),\n");
    fprintf(fp, "	.local_be (mem_local_be),\n");
    fprintf(fp, "	.local_init_done (ready),\n");
    fprintf(fp, "	.local_rdata (mem_local_rdata),\n");
    fprintf(fp, "	.local_rdata_valid (mem_local_rdata_valid),\n");
    fprintf(fp, "	.local_read_req (mem_local_read_req),\n");
    fprintf(fp, "	.local_ready (mem_local_ready),\n");
    fprintf(fp, "	.local_refresh_ack (),\n");
    fprintf(fp, "	.local_size (mem_local_size),\n");
    fprintf(fp, "	.local_wdata (mem_local_wdata),\n");
    fprintf(fp, "	.local_wdata_req (mem_local_wdata_req),\n");
    fprintf(fp, "	.local_write_req (mem_local_write_req),\n");
    fprintf(fp, "	.mem_addr (mem_addr),\n");
    fprintf(fp, "	.mem_ba (mem_ba),\n");
    fprintf(fp, "	.mem_cas_n (mem_cas_n),\n");
    fprintf(fp, "	.mem_cke (mem_cke),\n");
    fprintf(fp, "	.mem_clk (mem_clk),\n");
    fprintf(fp, "	.mem_clk_n (mem_clk_n),\n");
    fprintf(fp, "	.mem_cs_n (mem_cs_n),\n");
    fprintf(fp, "	.mem_dm (mem_dm),\n");
    fprintf(fp, "	.mem_dq (mem_dq),\n");
    fprintf(fp, "	.mem_dqs (mem_dqs),\n");
    fprintf(fp, "	.mem_odt (mem_odt),\n");
    fprintf(fp, "	.mem_ras_n (mem_ras_n),\n");
    fprintf(fp, "	.mem_we_n (mem_we_n),\n");
    if (mem_dqsn != 0)
    {
        fprintf(fp, "	.mem_dqsn (mem_dqsn),\n");
        fprintf(fp, "	.oct_ctl_rs_value (0),\n");
        fprintf(fp, "	.oct_ctl_rt_value (0),\n");
    }
    fprintf(fp, "	.phy_clk (phy_clk),\n");
    fprintf(fp, "	.pll_ref_clk (ref_clk),\n");
    fprintf(fp, "	.reset_phy_clk_n (reset_n),\n");
    fprintf(fp, "	.soft_reset_n (1)\n");
    fprintf(fp, "	);\n");
    fprintf(fp, "\n");
    fprintf(fp, "assign curReadAddrMem = ram_read_addr;\n");
    fprintf(fp, "assign curWriteByteEnMem = ram_write_byte_en;\n");
    fprintf(fp, "assign curWriteDataMem = ram_write_data;\n");
    fprintf(fp, "assign curWriteAddrMem = ram_write_addr;\n");
    fprintf(fp, "assign curWriteEnMem = ram_write_en && (left_sel == 0);\n");
    fprintf(fp, "assign leftWriteByteEnMem = ram_write_byte_en;\n");
    fprintf(fp, "assign leftWriteDataMem = ram_write_data;\n");
    fprintf(fp, "assign leftWriteAddrMem = ram_write_addr;\n");
    fprintf(fp, "assign leftWriteEnMem = ram_write_en && (left_sel == 1);\n");
    fprintf(fp, "assign ram_read_data = curReadDataMem;\n");

    fprintf(fp, "endmodule\n");
    fclose(fp);
}
