// generate wfifo module
// by Wei Zhang
//

#include <stdio.h>
#include <math.h>
#define intlog2(x) (int)ceil(log(x)/log(2))

int genwfifo(int fifosize, int memconwidth, int ratio)
{
	FILE *fp;
  int fifoinputwidth, fifooutputwidth;
  int fifosizewidth, fiforsizewidth;

  fifoinputwidth = memconwidth*ratio;
  fifooutputwidth = memconwidth;
  fifosizewidth = intlog2(fifosize);
  fiforsizewidth = fifosizewidth + intlog2(ratio);

	fp = fopen("wfifo.v", "w");

	fprintf(fp, "// synopsys translate_off\n");
	fprintf(fp, "`timescale 1 ps / 1 ps\n");
	fprintf(fp, "// synopsys translate_on\n");
	fprintf(fp, "module wfifo (\n");
	fprintf(fp, "	data,\n");
	fprintf(fp, "	rdclk,\n");
	fprintf(fp, "	rdreq,\n");
	fprintf(fp, "	wrclk,\n");
	fprintf(fp, "	wrreq,\n");
	fprintf(fp, "	q,\n");
	fprintf(fp, "	wrempty,\n");
	fprintf(fp, "	wrusedw);\n");
	fprintf(fp, "\n");
	fprintf(fp, "	parameter FIFOINPUTWIDTH = %i, FIFOSIZE = %i, FIFOSIZEWIDTH = %i;\n", fifoinputwidth, fifosize, fifosizewidth);
	fprintf(fp, "	parameter FIFOOUTPUTWIDTH = %i, FIFORSIZEWIDTH = %i;\n", fifooutputwidth, fiforsizewidth);
	fprintf(fp, "\n");
	fprintf(fp, "	input	[FIFOINPUTWIDTH-1:0]  data;\n");
	fprintf(fp, "	input	  rdclk;\n");
	fprintf(fp, "	input	  rdreq;\n");
	fprintf(fp, "	input	  wrclk;\n");
	fprintf(fp, "	input	  wrreq;\n");
	fprintf(fp, "	output	[FIFOOUTPUTWIDTH-1:0]  q;\n");
	fprintf(fp, "	output	  wrempty;\n");
	fprintf(fp, "	output	[FIFOSIZEWIDTH-1:0]  wrusedw;\n");
	fprintf(fp, "\n");
	fprintf(fp, "	wire [FIFOSIZEWIDTH-1:0] sub_wire0;\n");
	fprintf(fp, "	wire  sub_wire1;\n");
	fprintf(fp, "	wire [FIFOOUTPUTWIDTH-1:0] sub_wire2;\n");
	fprintf(fp, "	wire [FIFOSIZEWIDTH-1:0] wrusedw = sub_wire0[FIFOSIZEWIDTH-1:0];\n");
	fprintf(fp, "	wire  wrempty = sub_wire1;\n");
	fprintf(fp, "	wire [FIFOOUTPUTWIDTH-1:0] q = sub_wire2[FIFOOUTPUTWIDTH-1:0];\n");
	fprintf(fp, "\n");
	fprintf(fp, "	dcfifo_mixed_widths	dcfifo_mixed_widths_component (\n");
	fprintf(fp, "				.wrclk (wrclk),\n");
	fprintf(fp, "				.rdreq (rdreq),\n");
	fprintf(fp, "				.rdclk (rdclk),\n");
	fprintf(fp, "				.wrreq (wrreq),\n");
	fprintf(fp, "				.data (data),\n");
	fprintf(fp, "				.wrusedw (sub_wire0),\n");
	fprintf(fp, "				.wrempty (sub_wire1),\n");
	fprintf(fp, "				.q (sub_wire2)\n");
	fprintf(fp, "				// synopsys translate_off\n");
	fprintf(fp, "				,\n");
	fprintf(fp, "				.aclr (),\n");
	fprintf(fp, "				.rdempty (),\n");
	fprintf(fp, "				.rdfull (),\n");
	fprintf(fp, "				.rdusedw (),\n");
	fprintf(fp, "				.wrfull()\n");
	fprintf(fp, "				// synopsys translate_on\n");
	fprintf(fp, "				);\n");
	fprintf(fp, "	defparam\n");
	fprintf(fp, "		dcfifo_mixed_widths_component.intended_device_family = \"Stratix III\",\n");
	fprintf(fp, "		dcfifo_mixed_widths_component.lpm_numwords = FIFOSIZE,\n");
	fprintf(fp, "		dcfifo_mixed_widths_component.lpm_showahead = \"OFF\",\n");
	fprintf(fp, "		dcfifo_mixed_widths_component.lpm_type = \"dcfifo\",\n");
	fprintf(fp, "		dcfifo_mixed_widths_component.lpm_width = FIFOINPUTWIDTH,\n");
	fprintf(fp, "		dcfifo_mixed_widths_component.lpm_widthu = FIFOSIZEWIDTH,\n");
	fprintf(fp, "		dcfifo_mixed_widths_component.lpm_widthu_r = FIFORSIZEWIDTH,\n");
	fprintf(fp, "		dcfifo_mixed_widths_component.lpm_width_r = FIFOOUTPUTWIDTH,\n");
	fprintf(fp, "		dcfifo_mixed_widths_component.overflow_checking = \"OFF\",\n");
	fprintf(fp, "		dcfifo_mixed_widths_component.rdsync_delaypipe = 5,\n");
	fprintf(fp, "		dcfifo_mixed_widths_component.underflow_checking = \"OFF\",\n");
	fprintf(fp, "		dcfifo_mixed_widths_component.use_eab = \"ON\",\n");
	fprintf(fp, "		dcfifo_mixed_widths_component.wrsync_delaypipe = 5;\n");
	fprintf(fp, "\n");
	fprintf(fp, "\n");
	fprintf(fp, "endmodule\n");

	return 0;
}
